#!/usr/bin/env python

import sys
import argparse
import re

from dark.taxonomy import (
    formatLineage,
    addTaxonomyDatabaseCommandLineOptions,
    parseTaxonomyDatabaseCommandLineOptions,
)

VERSION_REGEX = re.compile(r"\.\d+$")


def taxonomyInfo(id_, db, namesOnly, separator):
    """
    Convert the lineage information for C{id_} into a formatted string.

    @param id_: The C{str} accession number or C{int} taxonomy id to
        look up.
    @param db: The sqlite3 taxonomy database to consult.
    @param namesOnly: If C{True} only print taxonomic names.
    @param separator: The C{str} separator to put between fields. If C{None}
        return space-padded aligned columns.
    @return: A formatted C{str} for printing or C{None} if C{id_} is
        not found in the taxonomy database.
    """
    try:
        idInt = int(id_)
    except ValueError:
        lineage = db.lineage(id_)
    else:
        lineage = db.lineage(idInt)

    if lineage is None and VERSION_REGEX.search(id_) is None:
        # Try adding a version number.
        lineage = db.lineage(id_ + ".1")

    if lineage:
        return formatLineage(lineage, namesOnly, separator)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Print taxonomy information for accession numbers or " "taxonomy ids."
        ),
    )

    parser.add_argument(
        "ids",
        nargs="*",
        metavar="id",
        help=(
            "The ids (accession numbers, names, or taxonomy ids) to print "
            "taxonomy information for. If not given, ids are read from "
            "standard input, one per line."
        ),
    )

    parser.add_argument(
        "--separator",
        help=(
            "The string to separate output columns with. Default is to "
            "print aligned output padded with multiple spaces (or a TAB if "
            "--namesOnly is used)."
        ),
    )

    parser.add_argument(
        "--namesOnly",
        default=False,
        action="store_true",
        help="If specified, only print the taxonomic names.",
    )

    parser.add_argument(
        "--print",
        default=False,
        action="store_true",
        help="If specified, also print the id.",
    )

    addTaxonomyDatabaseCommandLineOptions(parser)

    args = parser.parse_args()

    db = parseTaxonomyDatabaseCommandLineOptions(args, parser)

    if args.ids:
        ids = args.ids
    else:
        ids = (line.strip() for line in sys.stdin)

    separator = args.separator

    if args.namesOnly and separator is None:
        separator = "\t"

    for id_ in ids:
        if args.print:
            print(id_ + ":")
        result = taxonomyInfo(id_, db, args.namesOnly, separator)
        if result:
            print(result)
        else:
            print("%r was not found in the taxonomy database" % id_, file=sys.stderr)



#!/usr/bin/env python

import sys
import argparse
from os.path import exists
from os import mkdir, rename
from math import log10
from pathlib import Path

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions, Reads
from dark.utils import take


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=(
        "Split sequences in a FASTA/Q file into numerically named files, "
        "each containing a given number of sequences."
    ),
)

parser.add_argument("--outDir", default=".", help="The directory to make the files in.")

parser.add_argument(
    "--verbose",
    action="store_true",
    help="If given, print sequence ids as they are processed.",
)

parser.add_argument(
    "--count",
    type=int,
    default=100,
    help="The number of sequences to put in each file.",
)

parser.add_argument(
    "--noLeadingZeroes",
    action="store_true",
    help="If given, numeric filenames will not have leading zeroes.",
)

parser.add_argument(
    "--force", action="store_true", help="If given, overwrite pre-existing files."
)

parser.add_argument(
    "--saveAs",
    choices=("fasta", "fastq", "fasta-ss"),
    help=(
        "The output format. The default is to match the input format, "
        "so there is usually no need to specify this option. It can be "
        "used to force conversion from FASTQ to FASTA"
    ),
)

addFASTACommandLineOptions(parser)
args = parser.parse_args()
reads = parseFASTACommandLineOptions(args)

if not exists(args.outDir):
    mkdir(args.outDir)

saveAs = (
    args.saveAs
    or (args.fasta and "fasta")
    or (args.fastq and "fastq")
    or (args.fasta_ss and "fasta-ss")
)

# Note: we may be reading the FASTA input from stdin, so we cannot read it
# more than once (and I don't want to store it all because it may be very
# large). That's why we do a second phase of processing to renumber the
# files we created (if --noLeadingZeroes is not used).

outDir = Path(args.outDir)

count = 0
for count, reads in enumerate(
    take(parseFASTACommandLineOptions(args), args.count), start=1
):
    filename = outDir / f"{count}.fasta"
    if not args.force and filename.exists():
        print(
            f"Will not overwrite pre-existing file {str(filename)!r}. "
            f"Use --force to make me. Exiting.",
            file=sys.stderr,
        )
        sys.exit(1)
    if args.verbose:
        print(f"Writing {filename}")
    with open(filename, "w") as fp:
        Reads(reads).save(fp, saveAs)

# Rename numeric filenames to have leading zeroes.
if count and not args.noLeadingZeroes:
    width = int(log10(count)) + 1
    if width > 1:
        if args.verbose:
            print("Renaming to add leading zeroes.")
        for i in range(1, count + 1):
            old = str(i)
            new = "%0*d" % (width, i)
            if old == new:
                # We've reached the point where the new names have as many
                # digits as the old, so we can stop.
                break
            else:
                oldFilename = outDir / f"{old}.fasta"
                newFilename = outDir / f"{new}.fasta"
                if newFilename.exists() and not args.force:
                    print(
                        f"Will not overwrite pre-existing file "
                        f"{str(newFilename)!r}. Use --force to make me. "
                        f"Exiting.",
                        file=sys.stderr,
                    )
                    sys.exit(1)
                if args.verbose:
                    print(f"  {str(oldFilename)} -> {str(newFilename)}")
                rename(oldFilename, newFilename)



#!/usr/bin/env python

from typing import Dict

import sys
import argparse
from collections import defaultdict
from operator import itemgetter

from dark.aligners import edlibAlign, mafft, needle
from dark.dna import compareDNAReads
from dark.filter import (
    addFASTAFilteringCommandLineOptions,
    parseFASTAFilteringCommandLineOptions,
    addFASTAEditingCommandLineOptions,
    parseFASTAEditingCommandLineOptions,
)
from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

MAFFT_DEFAULT_ARGS = "--globalpair --maxiterate 1000 --preservecase"
MAFFT_ALGORITHMS_URL = (
    "https://mafft.cbrc.jp/alignment/software/algorithms/algorithms.html"
)
NEEDLE_DEFAULT_ARGS = "auto"


def align(reads, args):
    """
    Align a pair of reads.

    @param reads: An iterable of two C{DNARead} instances.
    @param args: An argparse C{Namespace} instance with command-line options.
    @return: A C{list} of two aligned C{DNARead} instances.
    """
    if args.aligner == "mafft":
        # Be careful in examining args.alignerOptions because we want the
        # user to be able to pass an empty string (so check against None
        # before deciding to use the default.)
        options = (
            MAFFT_DEFAULT_ARGS if args.alignerOptions is None else args.alignerOptions
        )
        return mafft(reads, args.verbose, options=options, threads=args.threads)
    elif args.aligner == "needle":
        # Be careful in examining args.alignerOptions because we want the
        # user to be able to pass an empty string (so check against None
        # before deciding to use the default.)
        options = (
            NEEDLE_DEFAULT_ARGS if args.alignerOptions is None else args.alignerOptions
        )
        return needle(reads, args.verbose, options=options)
    else:
        assert args.aligner == "edlib"
        return edlibAlign(reads)


def thresholdToCssName(threshold):
    """
    Turn a floating point threshold into a string that can be used as a CSS
    class name.

    param threshold: The C{float} threshold.
    return: A C{str} CSS class name.
    """
    # '.' is illegal in a CSS class name.
    return "threshold-" + str(threshold).replace(".", "_")


def thresholdForIdentity(identity, colors):
    """
    Get the best identity threshold for a specific identity value.

    @param identity: A C{float} nucleotide identity.
    @param colors: A C{list} of (threshold, color) tuples, where threshold is a
        C{float} and color is a C{str} to be used as a cell background. This
        is as returned by C{parseColors}.
    @return: The first C{float} threshold that the given identity is at least
        as big as.
    """
    for threshold, _ in colors:
        if identity >= threshold:
            return threshold
    raise ValueError("This should never happen! Last threshold is not 0.0?")


def parseColors(colors, defaultColor):
    """
    Parse command line color information.

    @param colors: A C{list} of space separated "value color" strings, such as
        ["0.9 red", "0.75 rgb(23, 190, 207)", "0.1 #CF3CF3"].
    @param defaultColor: The C{str} color to use for cells that do not reach
        the identity fraction threshold of any color in C{colors}.
    @return: A C{list} of (threshold, color) tuples, where threshold is a
        C{float} (from C{colors}) and color is a C{str} (from C{colors}). The
        list will be sorted by decreasing threshold values.
    """
    result = []
    if colors:
        for colorInfo in colors:
            fields = colorInfo.split(None, 1)
            if len(fields) == 2:
                threshold, color = fields
                try:
                    threshold = float(threshold)
                except ValueError:
                    print(
                        "--color arguments must be given as space-separated "
                        'pairs of "value color" where the value is a '
                        "numeric identity threshold. Your value %r is not "
                        "numeric." % threshold,
                        file=sys.stderr,
                    )
                    sys.exit(1)
                if 0.0 > threshold > 1.0:
                    print(
                        "--color arguments must be given as space-separated "
                        'pairs of "value color" where the value is a '
                        "numeric identity threshold from 0.0 to 1.0. Your "
                        "value %r is not in that range." % threshold,
                        file=sys.stderr,
                    )
                    sys.exit(1)

                result.append((threshold, color))
            else:
                print(
                    "--color arguments must be given as space-separated "
                    'pairs of "value color". You have given %r, which does '
                    "not contain a space." % colorInfo,
                    file=sys.stderr,
                )
                sys.exit(1)

    result.sort(key=itemgetter(0), reverse=True)

    if not result or result[-1][0] > 0.0:
        result.append((0.0, defaultColor))

    return result


def getGapCounts(reads, gapChars):
    """
    Get the number of gap characters in all reads.

    @param reads: A C{Reads} instance.
    @param gapChars: A C{str} of sequence characters considered to be gaps.
    @return: A C{dict} keyed by read id, with C{int} gap counts.
    """
    gapChars = set(gapChars)
    result = {}
    for read in reads:
        result[read.id] = sum(character in gapChars for character in read.sequence)
    return result


def getNoCoverageCounts(reads, noCoverageChars):
    """
    Get the no coverage counts for all reads.

    @param reads: A C{Reads} instance.
    @param noCoverageChars: A C{str} of sequence characters that indicate
        no coverage.
    @return: A C{dict} keyed by read id, with C{int} number of no coverage
        counts.
    """
    noCoverageChars = set(noCoverageChars)
    result = {}
    for read in reads:
        result[read.id] = sum(
            character in noCoverageChars for character in read.sequence
        )
    return result


def explanation(matchAmbiguous, concise, showLengths, showGaps, showNoCoverage, showNs):
    """
    Make an explanation of the output HTML table.

    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct. Otherwise, we are strict
        and insist that only non-ambiguous nucleotides can contribute to the
        matching nucleotide count.
    @param concise: If C{True}, do not show match detail abbreviations.
    @param showLengths: If C{True}, include the lengths of sequences.
    @param showGaps: If C{True}, include the number of gaps in sequences.
    @param showNoCoverage: If C{True}, include the number of no coverage
        characters in sequences.
    @param showNs: If C{True}, include the number of N characters in sequences.
    @return: A C{str} of HTML.
    """
    result = [
        """
<h1>Sequence versus sequence identity table</h1>

<p>

The table cells below show the nucleotide identity fraction for the sequences
(<span class="best">like this</span> for the best value in each row). The
identity fraction numerator is the sum of the number of identical
    """
    ]

    if matchAmbiguous:
        result.append(
            "nucleotides plus the number of ambiguously matching " "nucleotides."
        )
    else:
        result.append("nucleotides.")

    result.append(
        """The denominator
is the length of the sequence <em>for the row</em>. Sequence gaps
are not included when calculating their lengths.

</p>
    """
    )

    if showLengths or showGaps or showNs or matchAmbiguous or not concise:
        result.append(
            """
<p>

Key to abbreviations:
  <ul>
    """
        )

        if showLengths:
            result.append("<li>L: sequence <strong>L</strong>ength.</li>")

        if showGaps:
            result.append("<li>G: number of <strong>G</strong>aps in " "sequence.</li>")

        if showNoCoverage:
            result.append(
                "<li>C: number of no <strong>C</strong>overage "
                "characters in sequence.</li>"
            )

        if showNs:
            result.append(
                "<li>N: number of fully-ambiguous "
                "<strong>N</strong> characters in sequence.</li>"
            )

        if not concise:
            result.append(
                "<li>IM: <strong>I</strong>dentical nucleotide "
                "<strong>M</strong>atches.</li>"
            )

        if matchAmbiguous:
            result.append(
                "<li>AM: <strong>A</strong>mbiguous nucleotide "
                "<strong>M</strong>atches.</li>"
            )

        if showGaps:
            result.append(
                """
    <li>GG: Gap/Gap matches (both sequences have gaps).</li>
    <li>G?: Gap/Non-gap mismatches (one sequence has a gap).</li>
"""
            )

        if showNoCoverage:
            result.append(
                """
    <li>CC: No coverage/No coverage (both sequences have no coverage).</li>
    <li>C?: No coverage (one sequence has no coverage).</li>
"""
            )

    result.append(
        """
    <li>NE: Non-equal nucleotide mismatches.</li>
  </ul>
</p>
"""
    )

    return "\n".join(result)


def dataCell(
    id1: str, id2: str, square: bool, readNumbers: Dict[str, int], upperOnly: bool
) -> bool:
    """
    Should a cell have a value computed for it, or is it empty?

    @param id1: The row read id.
    @param id2: The column read id.
    @param square: If C{True} we are making a square table of a set of
        sequences against themselves (in which case we show nothing on the
        diagonal).
    @param readNumbers: A C{dict} mapping read ids to row numbers (only
        used if square is C{True} (in which case reads1 is the same as reads2).
    @param upperOnly: If C{True}, only compute values for the upper diagonal.
    @return: C{True} if the cell should be filled in, else C{False}.
    """
    if id1 == id2 and square:
        return False

    if not upperOnly:
        return True

    return readNumbers[id1] < readNumbers[id2]


def collectData(
    reads1,
    reads2,
    square,
    matchAmbiguous,
    pairwiseAlign,
    verbose,
    upperOnly=False,
    gapChars="-",
    noCoverageChars=None,
):
    """
    Get pairwise matching statistics for two sets of reads.

    @param reads1: An C{dict} of C{str} read ids whose values are
        C{Read} instances. These will be the rows of the table.
    @param reads2: An C{dict} of C{str} read ids whose values are
        C{Read} instances. These will be the columns of the table.
    @param square: If C{True} we are making a square table of a set of
        sequences against themselves (in which case we show nothing on the
        diagonal).
    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct. Otherwise, we are strict
        and insist that only non-ambiguous nucleotides can contribute to the
        matching nucleotide count.
    @param pairwiseAlign: If C{True}, pairwise-align the sequences.
    @param verbose: If C{True}, print progress output.
    @param upperOnly: If C{True}, only compute values for the upper diagonal.
    @param gapChars: A C{str} of sequence characters considered to be gaps.
    @param noCoverageChars: A C{str} of sequence characters that indicate no
        coverage.
    """
    readNumbers = {}
    comparisons = 0
    for readNumber, id1 in enumerate(reads1):
        readNumbers[id1] = readNumber

    for id1 in reads1:
        for id2 in reads2:
            if dataCell(id1, id2, square, readNumbers, upperOnly):
                comparisons += 1

    result = defaultdict(dict)
    count = 0

    for id1, read1 in reads1.items():
        for id2, read2 in reads2.items():
            if dataCell(id1, id2, square, readNumbers, upperOnly):
                count += 1
                if pairwiseAlign:
                    r1, r2 = align([read1, read2], args)
                else:
                    r1, r2 = read1, read2
                if verbose:
                    print(
                        f"Comparing {count}/{comparisons} {id1!r} " f"and {id2!r}.",
                        file=sys.stderr,
                    )
                match = compareDNAReads(
                    r1,
                    r2,
                    matchAmbiguous=matchAmbiguous,
                    gapChars=gapChars,
                    noCoverageChars=noCoverageChars,
                )
                if not matchAmbiguous:
                    assert match["match"]["ambiguousMatchCount"] == 0
                # Record the lengths, since these may have changed due to
                # making the alignment.
                match["read1"]["length"] = len(r1)
                match["read2"]["length"] = len(r2)
                result[id1][id2] = result[id2][id1] = match

    return result, readNumbers


def computeIdentity(read1, read2, stats, matchAmbiguous, digits):
    """
    Compute nucleotide identity for two reads (as a fraction of the lowest
    number of relevant nucleotides in either read).

    @param read1: A C{Read} instance.
    @param read2: A C{Read} instance.
    @param stats: A C{dict} as returned by C{compareDNAReads}.
    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct. Otherwise, we are strict
        and insist that only non-ambiguous nucleotides can contribute to the
        matching nucleotide count.
    @param digits: The C{int} number of digits to round values to.
    """

    # Note that the strict identity may be higher or lower than the
    # ambiguous identity even though an ambiguous match sounds like it
    # should always be more lenient and therefore result in a better
    # match. While it is true that the raw number of matched characters
    # will always increase when strict=False, the fraction of matched
    # characters may go down.
    #
    # The strict value can be higher because read1 might have many
    # ambiguous characters but very few of them may match read2. In that
    # case the overall fraction of matching characters will be pulled down
    # from the strict fraction when the ambiguous are included.
    #
    # Similarly, read1 may have many ambiguous characters, all of which are
    # matched by read2 and this can pull the overall identity higher than
    # the strict identity.

    match = stats["match"]
    denominator = min(stats["read1"]["length"], stats["read2"]["length"]) - (
        match["gapGapMismatchCount"]
        + match["noCoverageCount"]
        + match["noCoverageNoCoverageCount"]
    )
    numerator = stats["match"]["identicalMatchCount"]
    if matchAmbiguous:
        numerator += stats["match"]["ambiguousMatchCount"]

    result = numerator / denominator
    assert result <= 1.0, f"{numerator} / {denominator} = {result}.\n{stats}"

    return round(result, digits)


def textTable(
    tableData,
    reads1,
    reads2,
    readNumbers,
    square,
    matchAmbiguous,
    gapChars,
    numberedColumns,
    upperOnly=False,
    digits=3,
    addZeroes=False,
):
    """
    Make a text table showing inter-sequence distances.

    @param tableData: A C{defaultdict(dict)} keyed by read ids, whose values
        are the dictionaries returned by compareDNAReads.
    @param reads1: An C{dict} of C{str} read ids whose values are
        C{Read} instances. These will be the rows of the table.
    @param reads2: An C{dict} of C{str} read ids whose values are
        C{Read} instances. These will be the columns of the table.
    @param readNumbers: A C{dict} mapping read ids to row numbers (only
        used if square is C{True} (in which case reads1 is the same as reads2).
    @param square: If C{True} we are making a square table of a set of
        sequences against themselves (in which case we show nothing on the
        diagonal).
    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct. Otherwise, we are strict
        and insist that only non-ambiguous nucleotides can contribute to the
        matching nucleotide count.
    @param gapChars: A C{str} of sequence characters considered to be gaps.
    @param numberedColumns: If C{True}, use (row) numbers for column names.
    @param upperOnly: If C{True}, only show values for the upper diagonal.
    @param digits: The C{int} number of digits to round identities to.b
    @param addZeroes: If C{True}, add trailing zeroes to identities so they
        all have the same width.
    """
    titles = ["ID"]
    if numberedColumns:
        titles.extend(str(i + 1) for i in range(len(reads2)))

        if upperOnly and numberedColumns:
            titles.pop(1)
            titles[-1] = list(reads2)[-1]
    else:
        titles.extend(reads2)

    print("\t".join(titles))

    for rowCount, (id1, read1) in enumerate(reads1.items(), start=1):
        if upperOnly and numberedColumns and rowCount == len(reads1):
            # We don't print the last row when only showing the upper
            # diagonal, because it will be empty. It's name will appear at
            # the top of the final column.
            continue
        prefix = f"{rowCount}: " if numberedColumns else ""
        print(f"{prefix}{id1}", end="")
        for id2, read2 in reads2.items():
            if readNumbers[id2] == 0 and square:
                # The whole first column will be empty if we're making a
                # square array.
                continue
            if dataCell(id1, id2, square, readNumbers, upperOnly):
                identity = computeIdentity(
                    read1, read2, tableData[id1][id2], matchAmbiguous, digits
                )

                if addZeroes:
                    print(f"\t{identity:.{digits}f}", end="")
                else:
                    print(f"\t{identity}", end="")
            else:
                print("\t", end="")
        print()


def htmlTable(
    tableData,
    reads1,
    reads2,
    square,
    readNumbers,
    matchAmbiguous,
    colors,
    concise=False,
    showLengths=False,
    showGaps=False,
    showNoCoverage=False,
    showNs=False,
    footer=False,
    div=False,
    gapChars="-",
    noCoverageChars=None,
    numberedColumns=False,
    upperOnly=False,
    digits=3,
    addZeroes=False,
    highlightBest=False,
):
    """
    Make an HTML table showing inter-sequence distances.

    @param tableData: A C{defaultdict(dict)} keyed by read ids, whose values
        are the dictionaries returned by compareDNAReads.
    @param reads1: An C{dict} of C{str} read ids whose values are
        C{Read} instances. These will be the rows of the table.
    @param reads2: An C{dict} of C{str} read ids whose values are
        C{Read} instances. These will be the columns of the table.
    @param readNumbers: A C{dict} mapping read ids to row numbers (only
        used if square is C{True} (in which case reads1 is the same as reads2).
    @param square: If C{True} we are making a square table of a set of
        sequences against themselves (in which case we show nothing on the
        diagonal).
    @param matchAmbiguous: If C{True}, count ambiguous nucleotides that are
        possibly correct as actually being correct. Otherwise, we are strict
        and insist that only non-ambiguous nucleotides can contribute to the
        matching nucleotide count.
    @param colors: A C{list} of (threshold, color) tuples, where threshold is a
        C{float} and color is a C{str} to be used as a cell background. This
        is as returned by C{parseColors}.
    @param concise: If C{True}, do not show match details.
    @param showLengths: If C{True}, include the lengths of sequences.
    @param showGaps: If C{True}, include the number of gaps in sequences.
    @param showNoCoverage: If C{True}, include the number of no coverage
        characters in sequences.
    @param showNs: If C{True}, include the number of N characters in
        sequences.
    @param footer: If C{True}, incude a footer row giving the same information
        as found in the table header.
    @param div: If C{True}, return an HTML <div> fragment only, not a full HTML
        document.
    @param gapChars: A C{str} of sequence characters considered to be gaps.
    @param noCoverageChars: A C{str} of sequence characters that indicate no
        coverage.
    @param numberedColumns: If C{True}, use (row) numbers for column names.
    @param upperOnly: If C{True}, only show values for the upper diagonal.
    @param digits: The C{int} number of digits to round identities to.b
    @param addZeroes: If C{True}, add trailing zeroes to identities so they
        all have the same width.
    @param highlightBest: If C{True}, highlight the best identity value
        in each row.
    @return: An HTML C{str} showing inter-sequence distances.
    """
    gaps1 = getGapCounts(reads1.values(), gapChars)
    gaps2 = getGapCounts(reads2.values(), gapChars)
    noCoverage1 = getNoCoverageCounts(reads1.values(), noCoverageChars)
    noCoverage2 = getNoCoverageCounts(reads2.values(), noCoverageChars)
    result = []
    append = result.append

    def writeHeader():
        # The header row of the table.
        append("    <tr>")
        append("    <td>&nbsp;</td>")
        for count, read2 in enumerate(reads2.values(), start=1):
            if count == 1 and square:
                # The first column will be empty, so skip it.
                continue
            append(
                '    <td class="title"><span class="name">%s</span>'
                % (
                    count
                    if (upperOnly and numberedColumns and count != len(reads2))
                    else read2.id
                )
            )
            if not square:
                if showLengths:
                    append("    <br>L:%d" % len(read2))
                if showGaps:
                    append("    <br>G:%d" % gaps2[read2.id])
                if showNoCoverage:
                    append("    <br>C:%d" % noCoverage2[read2.id])
                if showNs:
                    append("    <br>N:%d" % read2.sequence.count("N"))
            append("    </td>")
        append("    </tr>")

    if div:
        append("<div>")
    else:
        append("<!DOCTYPE HTML>")
        append("<html>")
        append("<head>")
        append('<meta charset="UTF-8">')
        append("</head>")
        append("<body>")

    append("<style>")
    append(
        """
        table {
            border-collapse: collapse;
        }
        table, td {
            border: 1px solid #ccc;
        }
        tr:hover {
            background-color: #f2f2f2;
        }
        td {
            vertical-align: top;
            font-size: 14px;
        }
        span.name {
            font-weight: bold;
        }
        span.best {
            font-weight: bold;
        }
        td.nt-identity {
            text-align: right;
        }
    """
    )

    # Add color style information for the identity thresholds.
    for threshold, color in colors:
        append(".%s { background-color: %s; }" % (thresholdToCssName(threshold), color))

    append("</style>")

    if not div:
        append(
            explanation(
                matchAmbiguous, concise, showLengths, showGaps, showNoCoverage, showNs
            )
        )
    append('<div style="overflow-x:auto;">')
    append("<table>")
    append("  <tbody>")

    # Pre-process to find the best identities in each sample row.
    bestIdentityForId = {}
    identities = defaultdict(dict)

    for id1, read1 in reads1.items():
        # Look for best identity for the sample.
        bestIdentity = -1.0
        for id2, read2 in reads2.items():
            if dataCell(id1, id2, square, readNumbers, upperOnly):
                identity = computeIdentity(
                    read1, read2, tableData[id1][id2], matchAmbiguous, digits
                )
                identities[id1][id2] = identity
                if identity > bestIdentity:
                    bestIdentity = identity

        bestIdentityForId[id1] = bestIdentity

    writeHeader()

    # The main body of the table.
    for rowCount, (id1, read1) in enumerate(reads1.items(), start=1):
        if upperOnly and numberedColumns and rowCount == len(reads1):
            # We don't print the last row when only showing the upper
            # diagonal, because it will be empty. It's name will appear at
            # the top of the final column.
            continue

        append("    <tr>")
        append(
            '      <td class="title"><span class="name">%s%s</span>'
            % (f"{rowCount}: " if numberedColumns else "", id1)
        )
        if showLengths:
            append("<br/>L:%d" % len(read1))
        if showGaps:
            append("<br/>G:%d" % gaps1[read1.id])
        if showNoCoverage:
            append("<br/>C:%d" % noCoverage1[read1.id])
        if showNs:
            append("<br/>N:%d" % read1.sequence.count("N"))
        append("</td>")
        for id2, read2 in reads2.items():
            if readNumbers[id2] == 0 and square:
                # The whole first column will be empty if we're making a
                # square array.
                continue

            if not dataCell(id1, id2, square, readNumbers, upperOnly):
                append("<td>&nbsp;</td>")
                continue

            identity = identities[id1][id2]

            append(
                '      <td class="nt-identity %s">'
                % thresholdToCssName(thresholdForIdentity(identity, colors))
            )

            # The maximum percent identity.
            if highlightBest and identity == bestIdentityForId[id1]:
                scoreStyle = ' class="best"'
            else:
                scoreStyle = ""

            if addZeroes:
                append(f"<span{scoreStyle}>{identity:.{digits}f}</span>")
            else:
                append(f"<span{scoreStyle}>{identity}</span>")

            if not concise:
                match = tableData[id1][id2]["match"]
                append("<br/>IM:%d" % match["identicalMatchCount"])

                if matchAmbiguous:
                    append("<br/>AM:%d" % match["ambiguousMatchCount"])

                if showGaps:
                    append(
                        "<br/>GG:%d<br/>G?:%d"
                        % (match["gapGapMismatchCount"], match["gapMismatchCount"])
                    )

                if showNoCoverage:
                    append(
                        "<br/>CC:%d<br/>C?:%d"
                        % (match["noCoverageCount"], match["noCoverageNoCoverageCount"])
                    )

                append("<br/>NE:%d" % match["nonGapMismatchCount"])
            append("      </td>")
        append("    </tr>")

    if footer:
        writeHeader()

    append("  </tbody>")
    append("</table>")
    append("</div>")

    if div:
        append("</div>")
    else:
        append("</body>")
        append("</html>")

    return "\n".join(result)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Print a FASTA sequence identity table."
    )

    parser.add_argument(
        "--text",
        action="store_true",
        help="If specified, just print a simple text table.",
    )

    parser.add_argument(
        "--strict",
        action="store_true",
        help="If given, do not allow ambiguous nucleotide symbols to match.",
    )

    parser.add_argument(
        "--numberedColumns",
        action="store_true",
        help=(
            "Use a sequence (row) number as the header of each column "
            "instead of the sequence id."
        ),
    )

    parser.add_argument(
        "--concise", action="store_true", help="If given, do not show match details."
    )

    parser.add_argument(
        "--digits",
        default=3,
        type=int,
        help="The number of digits to round identities to.",
    )

    parser.add_argument(
        "--addZeroes",
        action="store_true",
        help=(
            "If given, make sure all identities have the same number of "
            "digits shown by adding zeroes when needed."
        ),
    )

    parser.add_argument(
        "--showLengths",
        action="store_true",
        help="If given, show the lengths of sequences.",
    )

    parser.add_argument(
        "--showGaps",
        action="store_true",
        help="If given, show the number of gaps in sequences.",
    )

    parser.add_argument(
        "--showNoCoverage",
        action="store_true",
        help=("If given, show the number of no-coverage characters in " "sequences."),
    )

    parser.add_argument(
        "--showNs",
        action="store_true",
        help=(
            "If given, show the number of fully ambiguous N characters in " "sequences."
        ),
    )

    parser.add_argument(
        "--footer",
        action="store_true",
        help="If given, also show sequence ids at the bottom of the table.",
    )

    parser.add_argument(
        "--div",
        action="store_true",
        help=(
            "If given, print an HTML <div> fragment only, not a full HTML "
            "document (ignored if --text is used)."
        ),
    )

    parser.add_argument(
        "--fastaFile2",
        type=open,
        metavar="FILENAME",
        help=(
            "The name of a second FASTA input file. If no second FASTA "
            "file is given, sequences in the first FASTA file will be "
            "compared with each other."
        ),
    )

    parser.add_argument(
        "--gapChars",
        default="-",
        metavar="CHARS",
        help=(
            "The sequence characters that should be considered to be gaps. "
            "These characters will be ignored in computing sequence lengths "
            "and identity fractions."
        ),
    )

    parser.add_argument(
        "--noCoverageChars",
        metavar="CHARS",
        help=(
            "The sequence characters that indicate lack of coverage. "
            "These characters will be ignored in identity fractions."
        ),
    )

    parser.add_argument(
        "--defaultColor",
        default="white",
        help=(
            "The (background) color for cells. This will be used for all "
            "cells that do not otherwise have a color due to use of "
            "--color. This option is ignored if --text is given."
        ),
    )

    parser.add_argument(
        "--color",
        action="append",
        help=(
            "Specify cell background coloring. This option must be given as "
            'a space separated "value color" pair. The value is an identity '
            "fraction in [0..1] and the color is any color "
            "specification that can be given to CSS. This argument can be "
            'repeated. E.g., --color "0.9 red" --color "0.75 rgb(23, 190, '
            '207)" --color "0.1 #CF3CF3". Cells will be colored using the '
            "color of the highest identity fraction they satisfy. The "
            "default is to color all cells with the --defaultColor color. "
            "This option is ignored if --text is given."
        ),
    )

    parser.add_argument(
        "--align", action="store_true", help="Do pairwise alignment of all sequences."
    )

    parser.add_argument(
        "--upperOnly",
        action="store_true",
        help=(
            "Only show the upper diagonal (only valid when a single FASTA "
            "input file is given, resulting in a square all-against-all "
            "identity table)."
        ),
    )

    parser.add_argument(
        "--sort", action="store_true", help="Sort the input sequences by id."
    )

    parser.add_argument(
        "--aligner",
        default="edlib",
        choices=("edlib", "mafft", "needle"),
        help="The alignment algorithm to use.",
    )

    parser.add_argument(
        "--alignerOptions",
        help=(
            "Optional arguments to pass to the alignment algorithm. If the "
            'aligner is mafft, the default options are %r. If needle, "%s". '
            "Do not try to set the number of threads here - use the "
            "--threads argument instead. If you are using mafft, see %s "
            "for some possible option combinations."
            % (MAFFT_DEFAULT_ARGS, NEEDLE_DEFAULT_ARGS, MAFFT_ALGORITHMS_URL)
        ),
    )

    parser.add_argument(
        "--verbose", action="store_true", help="Print progress to standard error."
    )

    parser.add_argument(
        "--highlightBest",
        action="store_true",
        help="Highlight the best identity in each row.",
    )

    addFASTACommandLineOptions(parser)
    addFASTAFilteringCommandLineOptions(parser)
    addFASTAEditingCommandLineOptions(parser)

    args = parser.parse_args()

    colors = parseColors(args.color, args.defaultColor)
    # Sanity check - the last threshold must be zero.
    assert colors[-1][0] == 0.0

    reads = parseFASTAEditingCommandLineOptions(
        args,
        parseFASTAFilteringCommandLineOptions(args, parseFASTACommandLineOptions(args)),
    )

    # Collect the reads into a dict, keeping the insertion order, unless we
    # are told to sort.
    reads1 = {}
    for read in sorted(reads) if args.sort else reads:
        reads1[read.id] = read

    if args.fastaFile2:
        if args.upperOnly:
            print(
                "The --upperOnly option is not supported if give two FASTA "
                "input files.",
                file=sys.stderr,
            )
            sys.exit(1)

        square = False
        reads2 = {}
        # The next line is a total hack, to trick parseFASTACommandLineOptions
        # into reading a second FASTA file.
        args.fastaFile = args.fastaFile2
        for read in parseFASTAFilteringCommandLineOptions(
            args, parseFASTACommandLineOptions(args)
        ):
            reads2[read.id] = read
    else:
        square = True
        reads2 = reads1

    matchAmbiguous = not args.strict
    tableData, readNumbers = collectData(
        reads1,
        reads2,
        square,
        matchAmbiguous,
        args.align,
        args.verbose,
        args.upperOnly,
        args.gapChars,
        args.noCoverageChars,
    )

    if args.text:
        textTable(
            tableData,
            reads1,
            reads2,
            readNumbers,
            square,
            matchAmbiguous,
            args.gapChars,
            args.numberedColumns,
            upperOnly=args.upperOnly,
            digits=args.digits,
            addZeroes=args.addZeroes,
        )
    else:
        print(
            htmlTable(
                tableData,
                reads1,
                reads2,
                square,
                readNumbers,
                matchAmbiguous,
                colors=colors,
                concise=args.concise,
                showLengths=args.showLengths,
                showGaps=args.showGaps,
                showNs=args.showNs,
                footer=args.footer,
                div=args.div,
                gapChars=args.gapChars,
                showNoCoverage=args.showNoCoverage,
                noCoverageChars=args.noCoverageChars,
                numberedColumns=args.numberedColumns,
                upperOnly=args.upperOnly,
                digits=args.digits,
                addZeroes=args.addZeroes,
                highlightBest=args.highlightBest,
            )
        )



#!/usr/bin/env python

import sys
import argparse
from numpy import std
from collections import defaultdict
from json import dumps

from dark.filter import (
    addFASTAFilteringCommandLineOptions,
    parseFASTAFilteringCommandLineOptions,
)
from dark.reads import Reads
from dark.sam import samfile, SAMFilter, samReferences, UnknownReference
from dark.utils import baseCountsToStr, pct

from gb2seq import Gb2SeqError
from gb2seq.alignment import Gb2Alignment, addAlignerOption
from gb2seq.features import Features, addFeatureOptions


def makeParser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Print SAM/BAM file coverage statistics by offset. "
            "Output lines show the offset."
        ),
    )

    addFASTAFilteringCommandLineOptions(parser)
    SAMFilter.addFilteringOptions(parser, samfileIsPositional=True)

    parser.add_argument(
        "--noOffsets",
        action="store_true",
        help="Do not print per-offset details of base counts.",
    )

    parser.add_argument(
        "--noStats",
        action="store_true",
        help="Do not print final average and standard deviation statistics.",
    )

    parser.add_argument(
        "--noFilter",
        action="store_true",
        help=(
            "Do not use our SAM filtering. Note that if you give this option, "
            "any filtering option (other than --referenceId) you also specify "
            "that is provided by the SAMFilter.addFilteringOptions will be "
            "silently ignored!"
        ),
    )

    addAlignerOption(parser)
    addFeatureOptions(
        parser,
        referenceHelpInfo=(
            " If you use this option, output will be printed as one "
            "JSON object per line, giving information not just about "
            "base coverage at each site but also about features in "
            "the reference and genome."
        ),
    )

    return parser


def printSiteInfo(
    alignment, site, includeUntranslated, includeGenome, bases, baseCount
):
    """
    Report what's found at a site for a given genome (or report insufficient
    coverage to standard error).

    @param alignment: A C{Gb2Alignment} instance.
    @param args: A C{Namespace} instance as returned by argparse with
        values for command-line options.
    @param includeGenome: If C{True}, include information about the genome
        (not just the reference).
    @param bases: A C{dict} mapping C{str} base names to C{int} counts.
    @param baseCount: The C{int} number of bases at the site (i.e., the
        coverage).
    """
    try:
        offsetInfo = alignment.offsetInfo(
            site,
            relativeToFeature=False,
            aa=False,
            featureName=None,
            includeUntranslated=includeUntranslated,
            allowAmbiguous=True,
        )
    except Gb2SeqError as e:
        print(e, file=sys.stderr)
        sys.exit(1)

    if offsetInfo is None:
        print("No coverage")
        return

    if not includeGenome:
        del offsetInfo["genome"]

    offsetInfo["featureNames"] = sorted(offsetInfo["featureNames"])

    # Add an "nt" entry to the reference and genome and delete the id.
    gappedOffset = alignment.gappedOffsets[site]
    for what, gappedSequence in zip(
        ("reference", "genome"),
        (alignment.referenceAligned.sequence, alignment.genomeAligned.sequence),
    ):
        info = offsetInfo[what]
        try:
            info["nt"] = gappedSequence[gappedOffset]
        except IndexError:
            error = (
                f"Index error determining nt at site {site} of {what}: "
                f"{gappedOffset=}. Alignment has length {len(gappedSequence)}. "
                f"Ref has length {len(alignment.features.reference)}."
            )
            # An index error was occurring before we started passing start,
            # stop, and truncate=True to pysam's pileup method, which was
            # returning an offset bigger than the reference genome
            # length. See
            # https://github.com/VirologyCharite/bih-pipeline/issues/199
            # (assuming you have access to that repo). This should now be
            # fixed. Here, just print a message to standard error, put the
            # error into the JSON that is printed, and keep going.
            print(error, file=sys.stderr)
            print(offsetInfo, file=sys.stderr)
            info["IndexError"] = error
        del offsetInfo[what]["id"]

    offsetInfo["bases"] = bases
    offsetInfo["coverage"] = baseCount

    print(dumps(offsetInfo, indent=None, sort_keys=True))


def main():
    parser = makeParser()
    args = parser.parse_args()

    if args.noOffsets and args.noStats:
        print(
            "You have used both --noOffsets and --noStats, so there is no " "output!",
            file=sys.stderr,
        )
        sys.exit(1)

    # We don't have a file of reads, we just want a read filter that we can use
    # to filter the SAM file query sequences and to get reference lengths from.
    reads = parseFASTAFilteringCommandLineOptions(args, Reads())
    samFilter = SAMFilter.parseFilteringOptions(args, reads.filterRead)

    printOffsets = not args.noOffsets
    printStats = not args.noStats

    if samFilter.referenceIds and len(samFilter.referenceIds) > 1:
        print(
            "Only one reference id can be given. To calculate coverage for more "
            "than one reference, run this script multiple times.",
            file=sys.stderr,
        )
        sys.exit(1)

    try:
        referenceLengths = samFilter.referenceLengths()
    except UnknownReference:
        referenceId = samFilter.referenceIds.pop()
        referenceIds = samReferences(args.samfile)
        print(
            "Reference %r does not appear in SAM file %s. Known "
            "references are: %s."
            % (referenceId, args.samfile, ", ".join(sorted(referenceIds))),
            file=sys.stderr,
        )
        sys.exit(1)

    if args.noFilter:
        # Do not do our custom SAM filtering.
        def filterRead(read):
            return not (read.is_del or read.is_refskip)

    else:

        def filterRead(read):
            return not (read.is_del or read.is_refskip) and samFilter.filterAlignment(
                read.alignment
            )

    if printStats:
        counts = []

    if args.reference or args.sars2:
        features = Features(
            args.reference,
            sars2=args.sars2,
            addUnannotatedRegions=args.addUnannotatedRegions,
        )

        alignment = Gb2Alignment(features.reference, features, aligner=args.aligner)
    else:
        alignment = None

    with samfile(args.samfile) as sam:
        if samFilter.referenceIds:
            # No need to check if the given reference id is in referenceLengths
            # because the samFilter.referenceLengths call above catches that.
            referenceId = samFilter.referenceIds.pop()
        else:
            if len(referenceLengths) == 1:
                referenceId = list(referenceLengths)[0]
            else:
                print(
                    "SAM file %r contains %d references (%s). Only one "
                    "reference id can be analyzed at a time. Please use "
                    "--referenceId to specify the one you want examined."
                    % (
                        args.samfile,
                        len(referenceLengths),
                        ", ".join(sorted(referenceLengths)),
                    ),
                    file=sys.stderr,
                )
                sys.exit(1)

        for column in sam.pileup(
            reference=referenceId,
            start=0,
            stop=referenceLengths[referenceId],
            truncate=True,
        ):
            bases = defaultdict(int)
            for read in column.pileups:
                if filterRead(read):
                    base = read.alignment.query_sequence[read.query_position]
                    bases[base] += 1

            baseCount = sum(bases.values())

            if printStats:
                counts.append(baseCount)

            if printOffsets:
                if alignment:
                    printSiteInfo(
                        alignment,
                        column.reference_pos,
                        args.addUnannotatedRegions,
                        True,
                        bases,
                        baseCount,
                    )
                else:
                    print(
                        f"{column.reference_pos + 1} {baseCount} "
                        f"{baseCountsToStr(bases)}"
                    )

    if printStats:
        referenceLength = referenceLengths[referenceId]
        print("Reference id: %s" % referenceId)
        print("Reference length: %d" % referenceLength)
        print("Bases covered: %s" % pct(len(counts), referenceLength))
        print(
            "Min coverage depth: %d"
            % (0 if len(counts) < referenceLength else min(counts))
        )
        if counts:
            # Don't use Python3 default= option on max. Trying to keep Python 2
            # compatibility.
            print("Max coverage depth: %d" % max(counts))
        else:
            print("Max coverage depth: 0")
        print("Mean coverage depth: %.3f" % (sum(counts) / referenceLength))
        if counts:
            print("Coverage depth s.d.: %.3f" % std(counts))


if __name__ == "__main__":
    main()



from six import PY3

from Bio.SeqIO.QualityIO import FastqGeneralIterator  # type: ignore

from dark.reads import Reads, DNARead
from dark.utils import asHandle


class FastqReads(Reads):
    """
    Subclass of L{dark.reads.Reads} providing access to FASTQ reads.

    @param _files: Either a single C{str} file name or file handle, or a
        C{list} of C{str} file names and/or file handles. Each file or file
        handle must contain sequences in FASTQ format.
    @param readClass: The class of read that should be yielded by iter.
    """

    def __init__(self, _files, readClass=DNARead):
        self._files = _files if isinstance(_files, (list, tuple)) else [_files]
        self.readClass = readClass
        if PY3:
            super().__init__()
        else:
            Reads.__init__(self)

    def iter(self):
        """
        Iterate over the sequences in the files in self.files_, yielding each
        as an instance of the desired read class.
        """
        for _file in self._files:
            with asHandle(_file) as fp:
                # Use FastqGeneralIterator because it provides access to
                # the unconverted quality string (i.e., it doesn't try to
                # figure out the numeric quality values, which we don't
                # care about at this point).
                for sequenceId, sequence, quality in FastqGeneralIterator(fp):
                    yield self.readClass(sequenceId, sequence, quality)



import six
from unittest import TestCase

from dark.mutations import getAPOBECFrequencies, mutateString


class TestGetAPOBECFrequencies(TestCase):
    """
    Tests the getAPOBECFrequencies function
    """

    def testCorrectResult(self):
        """
        The function should return the right bases.
        s:  AGTCTAAAGTCATGACTGGTCCCCTTG
        q:      ..AA..T....G....G...
        """
        result = []
        subjectTitle = "333031"
        subjectSequence = "AGTCTAAAGTCATGACTGGTCCCCTTG"
        subject = "%s \t \t \t %s" % (subjectTitle, subjectSequence)
        result.append(subject)

        queryTitle = "H9L0NJQ01DXGOQ"
        querySequence = "    ..AA..T....G....G...   "
        query = "%s \t %s" % (queryTitle, querySequence)
        result.append(query)

        bases = getAPOBECFrequencies(result, "C", "G", "cPattern")

        self.assertEqual(
            {
                "ACA": 0,
                "ACC": 0,
                "ACG": 0,
                "ACT": 1,
                "CCA": 0,
                "CCC": 0,
                "CCG": 0,
                "CCT": 0,
                "GCA": 0,
                "GCC": 0,
                "GCG": 0,
                "GCT": 0,
                "TCA": 0,
                "TCC": 1,
                "TCG": 0,
                "TCT": 0,
            },
            bases,
        )


class TestMutateString(TestCase):
    """
    Tests the mutateString function.
    """

    def testZeroLength(self):
        """
        The function should raise when given an empty string
        """
        error = "Empty original string passed."
        with six.assertRaisesRegex(self, ValueError, error):
            mutateString("", 1)

    def testTooManyMutationsRequested(self):
        """
        The function should raise when asked to perform more mutations than
        there are in the original string.
        """
        error = "Cannot make 2 mutations in a string of length 1"
        with six.assertRaisesRegex(self, ValueError, error):
            mutateString("x", 2)

    def testDuplicateReplacementLetter(self):
        """
        The function should raise when given a replacement string that contains
        a duplicate letter.
        """
        error = "Replacement string contains duplicates"
        with six.assertRaisesRegex(self, ValueError, error):
            mutateString("x", 1, "aa")

    def testReplacementLengthOneAppearsInOriginal(self):
        """
        The function should raise when given a replacement string that contains
        just one letter if that letter also appears in the original.
        """
        error = "Impossible replacement"
        with six.assertRaisesRegex(self, ValueError, error):
            mutateString("x", 1, "x")

    def testZeroReplacements(self):
        """
        The function should return the original string when zero mutations
        are requested.
        """
        self.assertEqual("acgt", mutateString("acgt", 0, "x"))

    def testOneDeterministicReplacement(self):
        """
        The function should return the correct result when one mutation is
        requested and only one mutation is possible.
        """
        self.assertEqual("c", mutateString("a", 1, "ac"))

    def testFiveDeterminsticReplacements(self):
        """
        The function should return the correct result when five mutations are
        requested and only one outcome is possible.
        """
        self.assertEqual("ccccc", mutateString("aaaaa", 5, "ac"))

    def testOneReplacement(self):
        """
        The function should return all possible correct results (and only those
        results) when one mutation is requested and two outcomes are possible.

        NOTE: this test is a bit bad because it's non-deterministic.
        """
        possible = set(["ab", "ba"])
        seen = set()

        for _ in range(100):
            result = mutateString("aa", 1, "ab")
            self.assertTrue(result in possible)
            seen.add(result)
        self.assertEqual(seen, possible)

    def testTwoReplacements(self):
        """
        The function should return all possible correct results (and only those
        results) when two mutations are requested and four outcomes are
        possible.

        NOTE: this test is a bit bad because it's non-deterministic.
        """
        possible = set(["ab", "ba", "ac", "ca"])
        seen = set()

        for _ in range(100):
            result = mutateString("aa", 1, "abc")
            self.assertTrue(result in possible, "%s is not in %r" % (result, possible))
            seen.add(result)
        self.assertEqual(seen, possible)



from __future__ import annotations

import sqlite3
from six import string_types
from operator import attrgetter
from cachetools import LRUCache, cachedmethod
from json import dumps
import re
from collections import namedtuple
from os import environ
from typing import Optional, Iterable
import argparse

from dark.database import getDatabaseConnection

TAXONOMY_DATABASE_ENV_VAR = "DARK_MATTER_TAXONOMY_DATABASE"
TAXONOMY_DATABASE_COMMAND_LINE_OPTION = "taxonomyDatabase"

LineageElement = namedtuple("LineageElement", ("taxid", "name", "rank"))

# These are rank/name tuples that indicate that a lineage is from an RNA
# virus.
RNA_VIRUS_LINEAGE_ELEMENTS = set(
    (
        ("family", "Retroviridae"),
        ("family", "Pseudoviridae"),
        ("realm", "Riboviria"),
    )
)

FUNGUS_ONLY_VIRUS_REGEX = re.compile(r"\b(?:mycovirus)\b", re.I)

FUNGUS_ONLY_GENERA: set[str] = set()

FUNGUS_ONLY_FAMILIES = {
    "Chrysoviridae",  # Infects penicillum. E.g., NC_040738.1
    "Totiviridae",  # Actually infects fungi and protozoa.
}

PLANT_ONLY_VIRUS_REGEX = re.compile(
    r"\b(?:grapevine|mosaic|mottle|blotch virus|viroid|tobacco|"
    r"tombus|cherry virus|bean endornavirus|coffee ringspot virus|"
    r"Odontoglossum ringspot virus|hibiscus|Obuda pepper virus|"
    r"tobamovirus|"
    r"wheat stripe virus|Lettuce necrotic yellows virus|citrus yellow|"
    r"Maize fine streak virus|dwarf (fiji)?virus|maize stripe virus|"
    r"tomato brown rugose fruit virus|turnip vein-clearing virus|"
    r"chlorotic spot virus)\b",
    re.I,
)

PLANT_ONLY_GENERA = {
    "Emaravirus",  # In Bunyavirales.
    "Idaeovirus",  # Unassigned.
    "Ourmiavirus",  # In Botourmiaviridae.
    "Polemovirus",  # Unassigned.
    "Sobemovirus",  # Unassigned.
    "Tospovirus",  # (Misclassified?) in Peribunyaviridae (e.g., NC_040742.1)
    "Umbravirus",  # In Tombusviridae.
    "Varicosavirus",  # In Rhabdoviridae.
}

PLANT_ONLY_FAMILIES = {
    "Aspiviridae",  # (Formerly Ophioviridae) in Serpentovirales.
    "Avsunviroidae",  # Viroids.
    "Betaflexiviridae",  # In Tymovirales.
    "Bromoviridae",  # Unassigned.
    "Caulimoviridae",  # In Ortervirales.
    "Closteroviridae",  # Unassigned.
    "Geminiviridae",  # Unassigned.
    "Luteoviridae",  # Unassigned.
    "Nanoviridae",  # Unassigned.
    "Partitiviridae",  # Unassigned.
    "Pospiviroidae",  # Unassigned (viroids).
    "Secoviridae",  # In Picornavirales.
    "Tombusviridae",  # Unassigned.
    "Tospoviridae",  # In Bunyavirales.
    "Virgaviridae",  # Unassigned.
}


PLANT_ONLY_ORDERS = {"Tymovirales"}


class _UnknownNameError(Exception):
    "An unknown name was used."


class LineageFetcher:
    """
    Provide access to the NCBI taxonomy database so we can retrieve the lineage
    of title sequences hit by BLAST.
    """

    def __init__(self, db=None, cursor=None):
        self._db = db or getDatabaseConnection()
        self._cursor = cursor or self._db.cursor()
        self._cache = {}

    def lineage(self, title: str) -> list[tuple[int, str]]:
        """
        Get lineage information from the taxonomy database for a given title.

        @param title: A C{str} sequence title (e.g., from a BLAST hit). Of the
            form 'gi|63148399|gb|DQ011818.1| Description...'. It is the gi
            number (63148399 in this example) that is looked up in the taxonomy
            database.
        @return: A C{list} of the taxonomic categories of the title. Each list
            element is an (C{int}, C{str}) 2-tuple, giving a taxonomy id and
            a scientific name. The first element in the list will correspond to
            C{title}, and each successive element is the parent of the
            preceeding one. If no taxonomy is found, the returned list will be
            empty.
        """
        if title in self._cache:
            return self._cache[title]

        lineage = []
        gi = int(title.split("|")[1])
        query = "SELECT taxID FROM gi_taxid WHERE gi = %d" % gi

        try:
            while True:
                self._cursor.execute(query)
                taxID = self._cursor.fetchone()[0]
                if taxID == 1:
                    break
                query = "SELECT name FROM names WHERE taxId = %s" % taxID
                self._cursor.execute(query)
                scientificName = self._cursor.fetchone()[0]
                lineage.append((taxID, scientificName))
                # Move up to the parent.
                query = "SELECT parent_taxID FROM nodes WHERE taxID = %s" % taxID
        except TypeError:
            lineage = []

        self._cache[title] = lineage
        return lineage

    def close(self) -> None:
        """
        Close the database connection and render self invalid. Any subsequent
        re-use of self will raise an error.
        """
        self._cursor.close()
        self._db.close()
        self._cursor = self._db = self._cache = None


class Taxonomy:
    """
    Provide access to the NCBI taxonomy database so we can retrieve the
    taxonomy lineage corresponding to an accession number.

    @param dbFilenameOrConnection: Either a C{str} database filename or an
        open sqlite3 database. The database must contain taxonomy information
        with tables and columns named as in the building scripts used in
        https://github.com/acorg/ncbi-taxonomy-database
    """

    CACHE_SIZE = 10e6

    def __init__(self, dbFilenameOrConnection):
        if isinstance(dbFilenameOrConnection, string_types):
            self._db = sqlite3.connect(dbFilenameOrConnection)
            self._closeConnection = True
        else:
            self._db = dbFilenameOrConnection
            self._closeConnection = False
        self._lineageCache = LRUCache(maxsize=self.CACHE_SIZE)
        self._hostsCache = LRUCache(maxsize=self.CACHE_SIZE)
        self._plantVirusCache = LRUCache(maxsize=self.CACHE_SIZE)
        self._fungusVirusCache = LRUCache(maxsize=self.CACHE_SIZE)

    def lineageFromTaxid(self, taxid, skipFunc=None, stopFunc=None):
        """
        Get lineage information from the taxonomy database for a taxonomy id.

        @param taxid: An C{int} taxonomy id.
        @param skipFunc: A function that takes a C{LineageElement} instance
            and returns C{True} if this lineage item should be excluded from
            the returned C{tuple}.
        @param stopFunc: A function that takes a C{LineageElement} instance
            and returns C{True} if the lineage processing should be stopped.
            The final lineage element is still added to the returned value,
            unless a C{skipFunc} is passed that returns C{True} for it.
        @raise ValueError: If the taxonomy id cannot be found in the names
            table or if any taxonomy id cannot be found in the nodes table
            as we move up the hierarchy.
        @raise AttributeError: If C{self.close} has been called.
        @return: A C{tuple} of the taxonomic categories of the title. Each
            tuple element is a C{LineageElement}, giving a taxonomy id a
            (scientific) name, and the rank (species, genus, etc). The first
            element in the list corresponds to the passed C{taxid} and each
            successive element is the parent of the preceeding one. The top
            level of the taxonomy hierarchy (with taxid = 1) is not included
            in the returned list. Note that the returned C{tuple} may be empty,
            depending on C{skipFunc} and C{stopFunc}.
        """
        cursor = self._db.cursor()
        execute = cursor.execute
        fetchone = cursor.fetchone

        lineage = []

        while taxid != 1:
            execute("SELECT name FROM names WHERE taxid = ?", (taxid,))
            result = fetchone()
            if result is None:
                raise ValueError(
                    "Could not find taxonomy id %r in names table" % (taxid,)
                )
            name = result[0]
            execute("SELECT parent_taxid, rank FROM nodes WHERE taxid = ?", (taxid,))
            result = fetchone()
            if result is None:
                raise ValueError(
                    "Could not find taxonomy id %r in nodes table" % (taxid,)
                )

            parentTaxid, rank = result
            element = LineageElement(taxid, name, rank)

            if not (skipFunc and skipFunc(element)):
                lineage.append(element)

            if stopFunc and stopFunc(element):
                break

            taxid = parentTaxid

        return tuple(lineage)

    @cachedmethod(attrgetter("_lineageCache"))
    def lineage(self, id_, skipFunc=None, stopFunc=None):
        """
        Get lineage information from the taxonomy database for an
        accession number, name, or taxonomy id.

        @param id_: A C{str} accession number, C{str} name (e.g.,
            'Hymenoptera'), or an C{int} taxonomy id. If a C{str} accession
            number, C{id_} must include the version (e.g., the ".1" part of
            "DQ011818.1") if the taxonomy database in use uses it (as is the
            case with the database built by
            https://github.com/acorg/ncbi-taxonomy-database)
        @param skipFunc: A function that takes a C{LineageElement} instance
            and returns C{True} if this lineage item should be excluded from
            the returned C{tuple}.
        @param stopFunc: A function that takes a C{LineageElement} instance
            and returns C{True} if the lineage processing should be stopped.
            The final lineage element is still added to the returned value,
            unless a C{skipFunc} is passed that returns C{True} for it.
        @raise ValueError: If a taxonomy id (or one of its higher-level
            taxonomy ids) cannot be found.
        @raise AttributeError: If C{self.close} has been called.
        @return: A C{tuple} of the taxonomic categories of the title. Each
            tuple element is a C{LineageElement}, giving a taxonomy id a
            (scientific) name, and the rank (species, genus, etc). The first
            element in the tuple corresponds to the passed C{id_} and each
            successive element is the parent of the preceeding one. The top
            level of the taxonomy hierarchy (with taxid = 1) is not included
            in the returned tuple.
        """
        if isinstance(id_, int):
            return self.lineageFromTaxid(id_, skipFunc=skipFunc, stopFunc=stopFunc)

        cursor = self._db.cursor()

        cursor.execute("SELECT taxid FROM accession_taxid WHERE accession = ?", (id_,))
        row = cursor.fetchone()
        if row:
            return self.lineageFromTaxid(
                int(row[0]), skipFunc=skipFunc, stopFunc=stopFunc
            )

        try:
            taxid = self.taxIdFromName(id_)
        except _UnknownNameError:
            raise ValueError(
                "Could not find %r in accession_taxid or names tables" % id_
            )
        else:
            return self.lineageFromTaxid(taxid, skipFunc=skipFunc, stopFunc=stopFunc)

    @cachedmethod(attrgetter("_hostsCache"))
    def hostsFromTaxid(self, taxid):
        """
        Get host information from the taxonomy database for a taxonomy id.

        @param taxid: An C{int} taxonomy id.
        @raise AttributeError: If C{self.close} has been called.
        @return: A C{set} of C{str} host names. If no host information is found
            for C{taxid}, return C{None}.
        """
        cursor = self._db.cursor()
        cursor.execute("SELECT hosts FROM hosts WHERE taxid = ?", (taxid,))
        row = cursor.fetchone()
        if row:
            return set(row[0].split(","))

    def taxIdFromName(self, name):
        """
        Get a taxonomy id from a name.

        @param name: A C{str} taxonomy name. This must be a unique name, as
            found in the names.dmp file of the NCBI taxonomy names.dmp file.
        @raise _UnknownNameError: If a taxonomy id cannot be found for C{name}.
        @return: An C{int} taxonomy id.
        """
        cursor = self._db.cursor()
        cursor.execute("SELECT taxid FROM names WHERE name = ?", (name,))
        row = cursor.fetchone()
        if row:
            return int(row[0])

        raise _UnknownNameError("Name %r not found in names tables" % name)

    def hosts(self, id_):
        """
        Get host information from the taxonomy database for an accession number
        or taxonomy id.

        @param id_: A C{str} accession number or C{int} taxonomy id.
            If a C{str}, C{id_} must include the version (e.g., the
            ".1" part of "DQ011818.1") if the taxonomy database in use uses
            it (as is the case with the database built by
            https://github.com/acorg/ncbi-taxonomy-database
        @raise AttributeError: If C{self.close} has been called.
        @return: A C{set} of C{str} host names, or C{None} if no host
            information.
        """
        if isinstance(id_, int):
            return self.hostsFromTaxid(id_)

        cursor = self._db.cursor()

        cursor.execute("SELECT taxid FROM accession_taxid WHERE accession = ?", (id_,))
        row = cursor.fetchone()
        if row:
            return self.hostsFromTaxid(int(row[0]))

        cursor.execute("SELECT taxid FROM names WHERE name = ?", (id_,))
        row = cursor.fetchone()
        if row:
            return self.hostsFromTaxid(int(row[0]))

    @cachedmethod(attrgetter("_fungusVirusCache"))
    def isFungusOnlyVirus(self, lineage, title=None) -> bool:
        """
        Determine whether a lineage corresponds to a fungus-only virus (i.e.,
        a virus that only infects fungi hosts).

        @param lineage: An iterable of C{tuple}s of taxonomy id, scientific
            name, and rank as returned by C{Taxonomy.lineage}.
        @param title: If not C{None}, a C{str} title of the virus.
        @return: C{True} if the lineage corresponds to a fungus-only virus,
            C{False} otherwise.
        """
        for taxid, name, rank in lineage:
            if (
                rank == "family"
                and name in FUNGUS_ONLY_FAMILIES
                or rank == "genus"
                and name in FUNGUS_ONLY_GENERA
            ):
                return True

        if title and FUNGUS_ONLY_VIRUS_REGEX.search(title):
            return True

        # Do host taxonomy database lookups. Try to look up host
        # information at all taxonomy levels. This is (of course) slower,
        # but it (likely, as with plant viruses) results in more
        # fungus-only viruses being identified.
        for taxid, _, _ in lineage:
            if self.hosts(taxid) == {"fungi"}:
                return True

        return False

    @cachedmethod(attrgetter("_plantVirusCache"))
    def isPlantOnlyVirus(self, lineage, title=None) -> bool:
        """
        Determine whether a lineage corresponds to a plant-only virus (i.e.,
        a virus that only infects plant hosts).

        @param lineage: An iterable of C{tuple}s of taxonomy id, scientific
            name, and rank as returned by C{Taxonomy.lineage}.
        @param title: If not C{None}, a C{str} title of the virus.
        @return: C{True} if the lineage corresponds to a plant-only virus,
            C{False} otherwise.
        """
        for taxid, name, rank in lineage:
            if (
                rank == "order"
                and name in PLANT_ONLY_ORDERS
                or rank == "family"
                and name in PLANT_ONLY_FAMILIES
                or rank == "genus"
                and name in PLANT_ONLY_GENERA
            ):
                return True

        if title and PLANT_ONLY_VIRUS_REGEX.search(title):
            return True

        # Do host taxonomy database lookups. Try to look up host information at
        # all taxonomy levels. This is (of course) slower, but it does result
        # in more plant-only viruses being identified, e.g., with rank
        # 'unclassified' and name 'Partitiviridae'.
        for taxid, _, _ in lineage:
            if self.hosts(taxid) == {"plants"}:
                return True

        return False

    @staticmethod
    def subsetLineageByRanks(lineage, func):
        """
        Extract certain ranks from a lineage.

        @param lineage: An iterable of C{tuple}s of taxonomy id, scientific
            name, and rank as returned by C{Taxonomy.lineage}.
        @param func: A function that accepts a C{str} rank (which may be '-')
            and returns C{True} or C{False} according to whether the lineage
            element should be retained.
        @return: A filter object that yields C{tuple}s of taxonomy id,
            scientific name, and rank, where the elements are those that
            C{func} returns C{True}. The elements in the returned value will
            be in the order in which they are present in C{lineage}.
        """
        return filter(lambda thisLineage: func(thisLineage.rank), lineage)

    def close(self) -> None:
        """
        Close the database connection (if we opened it).
        """
        if self._closeConnection:
            self._db.close()
        # Set self._db to None so self.lineage will raise an AttributeError
        # exception if it is called again.
        self._db = None

    def __enter__(self) -> Taxonomy:
        return self

    def __exit__(self, excType, excValue, traceback) -> None:
        self.close()


def isRetrovirus(lineage: Iterable[LineageElement]) -> bool:
    """
    Determine whether a lineage corresponds to a retrovirus.

    @param lineage: An iterable of C{LineageElement} instances.
    @return: C{True} if the lineage corresponds to a retrovirus, C{False}
        otherwise.
    """
    for element in lineage:
        if element.rank == "family" and element.name == "Retroviridae":
            return True

    return False


def isRNAVirus(lineage):
    """
    Determine whether a lineage corresponds to an RNA virus.

    @param lineage: An iterable of C{LineageElement} instances.
    @return: C{True} if the lineage corresponds to an RNA virus, C{False}
        otherwise.
    """
    for element in lineage:
        if (element.rank, element.name) in RNA_VIRUS_LINEAGE_ELEMENTS:
            return True

    return False


def isDNAVirus(lineage: Iterable[LineageElement]) -> bool:
    """
    Determine whether a lineage corresponds to an DNA virus.

    @param lineage: An iterable of C{LineageElement} instances.
    @return: C{True} if the lineage corresponds to a DNA virus, C{False}
        otherwise.
    """
    return not isRNAVirus(lineage)


def formatLineage(
    lineage, namesOnly: bool = False, separator: str = ", ", prefix: str = ""
) -> str:
    """
    Format a lineage for printing.

    @param lineage: An iterable of C{LineageElement} instances.
    @param namesOnly: If C{True} only print taxonomic names.
    @param separator: A C{str} separator to put between fields. If C{None},
        return a space-padded aligned columns.
    @param prefix: A C{str} to put at the start of each line.
    @return: A formatted C{str} for printing.
    """
    if namesOnly:
        # The separator is guaranteed to be set by our caller.
        return prefix + separator.join(element.name for element in lineage)

    if separator is not None:
        return "\n".join(
            "%s%s%s%s%s%s" % (prefix, rank, separator, name, separator, taxid)
            for (taxid, name, rank) in lineage
        )

    # This is a bit slow, walking through the lineage list 3 times.
    taxidWidth = max(len(str(element.taxid)) for element in lineage)
    nameWidth = max(len(element.name) for element in lineage)
    rankWidth = max(len(element.rank) for element in lineage)

    return "\n".join(
        "%s%-*s %-*s %*d"
        % (prefix, rankWidth, rank, nameWidth, name, taxidWidth, taxid)
        for (taxid, name, rank) in lineage
    )


def lineageTaxonomyLinks(lineage: Iterable[LineageElement]) -> list[str]:
    """
    Get HTML links for a lineage.

    @param lineage: An iterable of C{LineageElement} instances.
    @return: A C{list} of HTML C{str} links.
    """
    names = [element.name for element in lineage]
    names[0] = "taxon"

    taxids = [element.taxid for element in lineage]

    return [
        '<a href="%s%s">%s</a>'
        % ("https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=", taxid, name)
        for taxid, name in list(zip(taxids, names))[:-1]
    ]


class _HierarchyNode:
    def __init__(self, name: str):
        self.name = name
        self.count = 0
        self.nodes: dict[str, _HierarchyNode] = {}
        self.tips: list[tuple[str, str]] = []

    def addChild(self, name) -> _HierarchyNode:
        self.count += 1
        try:
            node = self.nodes[name]
        except KeyError:
            node = self.nodes[name] = _HierarchyNode(name)

        return node

    def addTip(self, name: str, genomeAccession: str) -> None:
        self.count += 1
        self.tips.append((name, genomeAccession))

    def toDict(self) -> dict:
        nodes = [node.toDict() for node in self.nodes.values()]

        if self.tips:
            nodes.extend(
                [
                    {
                        "text": name,
                        "href": "#pathogen-" + accession,
                    }
                    for (name, accession) in sorted(self.tips)
                ]
            )

        return {
            # 'count': self.count,
            "nodes": nodes,
            # 'name': self.name,
            # 'tips': self.tips,
            "text": "%s (%d)" % (self.name, self.count),
        }


class Hierarchy:
    """
    Collect information about a lineages in a taxonomy hierarchy.
    """

    def __init__(self, rootName: str = "Viruses"):
        self._root = _HierarchyNode(rootName)

    def add(self, lineage: Iterable[LineageElement], genomeAccession: str) -> None:
        """
        Add a new lineage.

        @param lineage: An iterable of C{LineageElement} instances.
        @param genomeAccession: A C{str} pathogen accession number.
        """
        names = [element.name for element in lineage]

        n = len(names) - 2
        node = self._root

        while n:
            node = node.addChild(names[n])
            n -= 1

        node.addTip(names[0], genomeAccession)

    def toJSON(self):
        return dumps([self._root.toDict()])


def addTaxonomyDatabaseCommandLineOptions(parser: argparse.ArgumentParser) -> None:
    """
    Add standard taxonomy database command-line options to an argparse parser.

    @param parser: An C{argparse.ArgumentParser} instance.
    """
    parser.add_argument(
        "--" + TAXONOMY_DATABASE_COMMAND_LINE_OPTION,
        metavar="DATABASE-FILE",
        help=(
            "The file holding an sqlite3 taxonomy database. See "
            "https://github.com/acorg/ncbi-taxonomy-database for how to "
            "build one. If not specified, the value in the %r environment "
            "variable (if any) will be used." % TAXONOMY_DATABASE_ENV_VAR
        ),
    )


def parseTaxonomyDatabaseCommandLineOptions(args, parser) -> Optional[Taxonomy]:
    """
    Examine parsed command-line options and return a Taxonomy instance. Exits
    if no taxonomy database is given or named in the TAXONOMY_DATABASE_ENV_VAR
    environment variable.

    @param args: An argparse namespace, as returned by the argparse
        C{parse_args} function.
    @param parser: An C{argparse.ArgumentParser} instance.
    @return: A C{Taxonomy} instance, or C{None} if no database filename is
        given or set via the environment variable named in
        TAXONOMY_DATABASE_ENV_VAR.
    """
    filename = args.taxonomyDatabase or environ.get(TAXONOMY_DATABASE_ENV_VAR)

    if filename:
        return Taxonomy(filename)
    else:
        parser.error(
            "A taxonomy database file must be given, either via --%s on "
            "the command line or via the %r environment variable."
            % (TAXONOMY_DATABASE_COMMAND_LINE_OPTION, TAXONOMY_DATABASE_ENV_VAR)
        )

        return None



#!/usr/bin/env python

import sys
import argparse
from collections import defaultdict

from dark.sam import samfile, SAMFilter
from dark.utils import pct


def referenceInfo():
    """
    Hold information about the reads that match a reference.

    @return: A C{dict}, as below.
    """
    return {
        "duplicate": set(),
        "primary": set(),
        "qcFail": set(),
        "nonDuplicate": set(),
        "readIds": set(),
        "secondary": set(),
        "supplementary": set(),
    }


def main(args):
    """
    Print SAM/BAM file reference read counts.

    @param args: An argparse namespace with information about parsed
        command-line options.
    """
    if args.topReferenceIdsFile and args.sortBy != "count":
        print(
            "--topReferenceIdsFile only makes sense when using --sortBy " "count",
            file=sys.stderr,
        )
        sys.exit(1)

    referenceReads = defaultdict(referenceInfo)
    mapped = set()
    unmapped = set()
    readIds = set()

    referenceLengths = SAMFilter(args.samFile).referenceLengths()

    with samfile(args.samFile) as fp:
        for read in fp.fetch():
            id_ = read.query_name
            readIds.add(id_)
            if read.is_unmapped:
                unmapped.add(id_)
            else:
                mapped.add(id_)
                stats = referenceReads[read.reference_name]
                stats["readIds"].add(id_)

                if read.is_secondary:
                    stats["secondary"].add(id_)
                elif read.is_supplementary:
                    stats["supplementary"].add(id_)
                else:
                    stats["primary"].add(id_)

                if read.is_duplicate:
                    stats["duplicate"].add(id_)
                else:
                    stats["nonDuplicate"].add(id_)

                if read.is_qcfail:
                    stats["qcFail"].add(id_)

    totalReads = len(readIds)

    print(
        "Found a total of %d read%s (%d mapped, %d unmapped) mapping "
        "against %d of %d reference%s."
        % (
            totalReads,
            "" if totalReads == 1 else "s",
            len(mapped),
            len(unmapped),
            len(referenceReads),
            len(referenceLengths),
            "" if len(referenceLengths) == 1 else "s",
        )
    )

    if args.sortBy == "count":

        def key(referenceId):
            return len(referenceReads[referenceId]["readIds"])

        sortedReferenceReads = sorted(referenceReads, key=key, reverse=True)
        topReference = sortedReferenceReads[0]
    else:
        # Sort the references by name
        sortedReferenceReads = sorted(referenceReads)

    cumulativeReadIds = set()

    for count, referenceId in enumerate(sortedReferenceReads, start=1):
        stats = referenceReads[referenceId]
        readCount = len(stats["readIds"])
        if readCount == 0 and args.excludeZeroes:
            continue
        newReadCount = len(stats["readIds"] - cumulativeReadIds)
        if newReadCount == 0 and args.excludeIfNoAdditional:
            continue
        cumulativeReadIds.update(stats["readIds"])
        print(
            "\nReference %d: %s (%d nt):\n"
            "  Overall reads mapped to the reference: %s\n"
            "  Non-duplicates: %s, Duplicates: %s, QC fails: %s\n"
            "  Primary: %s, Secondary: %s, Supplementary: %s\n"
            "  Reads not matching any reference above: %s\n"
            "  Previously unmatched reads for this reference: %s"
            % (
                count,
                referenceId,
                referenceLengths[referenceId],
                pct(readCount, totalReads),
                pct(len(stats["nonDuplicate"]), readCount),
                pct(len(stats["duplicate"]), readCount),
                pct(len(stats["qcFail"]), readCount),
                pct(len(stats["primary"]), readCount),
                pct(len(stats["secondary"]), readCount),
                pct(len(stats["supplementary"]), readCount),
                pct(newReadCount, totalReads),
                pct(newReadCount, readCount),
            )
        )

    # Write out the (sorted) read ids of the reference with the most reads.
    if args.topReferenceIdsFile:
        with open(args.topReferenceIdsFile, "w") as fp:
            print("\n".join(sorted(referenceReads[topReference]["readIds"])), file=fp)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Print SAM/BAM file reference sequence and unmapped read " "counts."
        ),
    )

    parser.add_argument(
        "samFile", metavar="FILENAME", help="The name of a SAM/BAM alignment file."
    )

    parser.add_argument(
        "--sortBy",
        choices=("count", "name"),
        default="count",
        help=(
            'Use "count" to sort the output by decreasing read count (i.e., '
            "the reference with the highest number of matching reads is "
            'printed first), or "name" to sort by reference name.'
        ),
    )

    parser.add_argument(
        "--excludeZeroes",
        action="store_true",
        help="Do not print references unmatched by any read.",
    )

    parser.add_argument(
        "--excludeIfNoAdditional",
        action="store_true",
        help=(
            "Do not print references that are mapped by no additional reads "
            "(i.e., reads other than those matching already-printed "
            "references). This is useful when --sortBy count is used, "
            "making it possible to not print references only matched by "
            "reads matching references that have already been printed."
        ),
    )

    parser.add_argument(
        "--topReferenceIdsFile",
        metavar="FILE",
        help=(
            "The file to write the (sorted) ids of the reads for the "
            "reference with the highest number of matching reads. Only "
            "valid when --sortBy count is used."
        ),
    )

    main(parser.parse_args())



#!/usr/bin/env python

import sys
import warnings
from time import time
from itertools import chain
import argparse

from dark.civ.proteins import SqliteIndexWriter
from dark.taxonomy import (
    addTaxonomyDatabaseCommandLineOptions,
    parseTaxonomyDatabaseCommandLineOptions,
)


def filenamesAndAdders(args, db):
    """
    Get the filenames and database adding functions.

    @param args: A C{Namespace} instance as returned by argparse.
    @param db: An C{SqliteIndexWriter} instance.
    @return: A generator yielding 3-tuples containing a filename, a
        database adder function that can read the file and incorporate it,
        and a C{str} that is either 'gb' or 'json'.
    """
    # Use chain to flatten the lists of lists that we get from using both
    # nargs='+' and action='append'. We use both because it allows people
    # to use (e.g.) --gb on the command line either via "--gb file1 --gb
    # file2" or "--gb file1 file2", or a combination of these. That way
    # it's not necessary to remember which way you're supposed to use it
    # and you also can't be hit by the subtle problem encountered in
    # https://github.com/acorg/dark-matter/issues/453

    if not (args.gb or args.json):
        print(
            "At least one of --gb or --json must be used to indicate input "
            "files to process.",
            file=sys.stderr,
        )
        sys.exit(1)

    if args.gb:
        for filename in chain.from_iterable(args.gb):
            yield filename, db.addGenBankFile, "gb"

    if args.json:
        for filename in chain.from_iterable(args.json):
            yield filename, db.addJSONFile, "json"


def main(args, parser):
    """
    Build the protein database.

    @param args: The namespace of command-line arguments returned by
        argparse.parse_args()
    @param parser: An C{argparse.ArgumentParser} instance.
    """

    if (
        args.minGenomeLength is not None
        and args.maxGenomeLength is not None
        and args.minGenomeLength > args.maxGenomeLength
    ):
        raise ValueError("--minGenomeLength cannot be larger than --maxGenomeLength")

    if args.excludeExclusiveHost:
        excludeExclusiveHosts = set(chain.from_iterable(args.excludeExclusiveHost))
    else:
        excludeExclusiveHosts = None

    taxonomyDatabase = parseTaxonomyDatabaseCommandLineOptions(args, parser)
    progress = args.progress

    if progress:
        overallStart = time()
        totalGenomeCount = totalProteinCount = 0

    with SqliteIndexWriter(args.databaseFile) as db:
        for fileCount, (filename, addFunc, type_) in enumerate(
            filenamesAndAdders(args, db), start=1
        ):
            if args.logFile:
                print("\n>>> Indexing '%s'." % filename, end="\n\n", file=args.logFile)

            if progress:
                start = time()

            examinedGenomeCount, genomeCount, proteinCount = addFunc(
                filename,
                dnaOnly=args.dnaOnly,
                rnaOnly=args.rnaOnly,
                minGenomeLength=args.minGenomeLength,
                maxGenomeLength=args.maxGenomeLength,
                excludeExclusiveHosts=excludeExclusiveHosts,
                excludeFungusOnlyViruses=args.excludeFungusOnlyViruses,
                excludePlantOnlyViruses=args.excludePlantOnlyViruses,
                databaseName=args.databaseName,
                taxonomyDatabase=taxonomyDatabase,
                proteinSource=args.proteinSource,
                genomeSource=args.genomeSource,
                duplicationPolicy=args.duplicationPolicy,
                logfp=args.logFile,
            )

            if examinedGenomeCount == 0:
                if type_ == "gb":
                    print(
                        "WARNING: No genomes found in %r. Did the GenBank "
                        "download fail on that file?" % filename,
                        file=sys.stderr,
                    )
                else:
                    assert type_ == "json"
                    print(
                        "WARNING: no genomes found in JSON file %r." % filename,
                        file=sys.stderr,
                    )

            if progress:
                elapsed = time() - start
                totalGenomeCount += genomeCount
                totalProteinCount += proteinCount
                print(
                    "Processed %r: added %3d of %3d genome%s (%5d "
                    "protein%s) in %.2f seconds."
                    % (
                        filename,
                        genomeCount,
                        examinedGenomeCount,
                        " " if examinedGenomeCount == 1 else "s",
                        proteinCount,
                        "" if proteinCount == 1 else "s",
                        elapsed,
                    ),
                    file=sys.stderr,
                )

    if progress:
        elapsed = time() - overallStart
        print(
            "%d files (containing %d genomes and %d proteins) "
            "indexed in %.2f seconds (%.2f mins)."
            % (fileCount, totalGenomeCount, totalProteinCount, elapsed, elapsed / 60),
            file=sys.stderr,
        )


parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description=(
        "Create a genome/protein sqlite3 database from GenBank "
        "and JSON files. The protein sequences for the database are "
        "printed to standard output for indexing by other tools."
    ),
)

parser.add_argument(
    "--databaseFile",
    required=True,
    help=("The output file. This file must not exist (use --force to " "overwrite)."),
)

parser.add_argument(
    "--databaseName",
    help=(
        "The database that the records in the (--gb) GenBank files came "
        'from (e.g., "refseq" or "RVDB").'
    ),
)

parser.add_argument(
    "--duplicationPolicy",
    choices=("error", "ignore"),
    default="ignore",
    help=(
        "What to do if a to-be-inserted accession number is already "
        'present in the database. "error" results in a ValueError being '
        'raised, "ignore" will ignore the duplicate. It should also be '
        "possible to update (i.e., replace) but that is not supported yet."
    ),
)

parser.add_argument("--progress", action="store_true", help="Print indexing progress.")

parser.add_argument(
    "--logFile",
    type=argparse.FileType("w"),
    help="Write indexing details to a log file.",
)

parser.add_argument(
    "--noWarnings",
    action="store_true",
    help="Do not print warnings about unparseable GenBank or JSON records.",
)

# A mutually exclusive group for DNA/RNA only.
group = parser.add_mutually_exclusive_group()

group.add_argument(
    "--dnaOnly", action="store_true", help="If given, only include DNA viruses."
)

group.add_argument(
    "--rnaOnly", action="store_true", help="If given, only include RNA viruses."
)

parser.add_argument(
    "--maxGenomeLength",
    type=int,
    help="Genomes longer than this will not be considered.",
)

parser.add_argument(
    "--minGenomeLength",
    type=int,
    help="Genomes shorter than this will not be considered.",
)

parser.add_argument(
    "--excludeFungusOnlyViruses",
    action="store_true",
    help=(
        "If given, exclude fungus-only viruses (i.e., viruses that only "
        "infect fungi host species)."
    ),
)

parser.add_argument(
    "--excludePlantOnlyViruses",
    action="store_true",
    help=(
        "If given, exclude plant-only viruses (i.e., viruses that only "
        "infect plant host species)."
    ),
)

parser.add_argument(
    "--gb",
    metavar="GenBank-file",
    nargs="+",
    action="append",
    help=(
        "The GenBank file(s) to make the database from. These may be "
        "uncompressed, or compressed with bgzip (from samtools), with "
        "a .gz suffix."
    ),
)

parser.add_argument(
    "--json",
    metavar="JSON-file",
    nargs="+",
    action="append",
    help=(
        "The JSON file(s) to make the database from. These contain "
        "genome and protein information for cases where sequences that "
        "are not in GenBank should be added."
    ),
)

parser.add_argument(
    "--proteinSource",
    default="GenBank",
    help=(
        "The source of the accession numbers for the proteins found in the "
        "input files. This becomes part of the sequence id printed in the "
        "protein FASTA output."
    ),
)

parser.add_argument(
    "--genomeSource",
    default="GenBank",
    help=(
        "The source of the accession numbers for the genomes in the input "
        "files. This becomes part of the sequence id printed in the "
        "protein FASTA output."
    ),
)

parser.add_argument(
    "--excludeExclusiveHost",
    metavar="hostname",
    nargs="+",
    action="append",
    choices=(
        "algae",
        "archaea",
        "bacteria",
        "diatom",
        "environment",
        "eukaryotes",
        "fungi",
        "human",
        "invertebrates",
        "plants",
        "protozoa",
        "vertebrates",
    ),
    help=(
        "A host type that should be excluded, but only if this is the only "
        "host of the virus."
    ),
)

addTaxonomyDatabaseCommandLineOptions(parser)

args = parser.parse_args()

if args.noWarnings:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        main(args, parser)
else:
    main(args, parser)



import six
from unittest import TestCase

from dark.reads import Read, Reads
from dark.score import HigherIsBetterScore
from dark.hsp import HSP, LSP
from dark.alignments import (
    Alignment,
    bestAlignment,
    ReadAlignments,
    ReadsAlignmentsParams,
    ReadsAlignments,
)


class TestAlignment(TestCase):
    """
    Tests for the dark.alignment.Alignment class
    """

    def testExpectedAttrs(self):
        """
        An alignment must have the expected attributes.
        """
        alignment = Alignment(45, "title")
        self.assertEqual("title", alignment.subjectTitle)
        self.assertEqual(45, alignment.subjectLength)

    def testNoHspsWhenCreated(self):
        """
        An alignment must have no HSPs when it is created.
        """
        alignment = Alignment(45, "title")
        self.assertEqual(0, len(alignment.hsps))

    def testAddHsp(self):
        """
        It must be possible to add an HSP to an alignment.
        """
        alignment = Alignment(45, "title")
        alignment.addHsp(HSP(3))
        self.assertEqual(HSP(3), alignment.hsps[0])


class TestReadAlignments(TestCase):
    """
    Tests for the dark.alignment.ReadAlignments class
    """

    def testRead(self):
        """
        An read alignments must store its read.
        """
        read = Read("id", "ACGT")
        readAlignments = ReadAlignments(read)
        self.assertEqual(read, readAlignments.read)

    def testNoAlignments(self):
        """
        An read alignments must be able to have no alignments.
        """
        read = Read("id", "ACGT")
        readAlignments = ReadAlignments(read)
        self.assertEqual(0, len(readAlignments))

    def testAlignments(self):
        """
        An read alignments must store its alignments.
        """
        read = Read("id", "ACGT")
        alignment1 = Alignment(45, "title1")
        alignment2 = Alignment(55, "title2")
        readAlignments = ReadAlignments(read, [alignment1, alignment2])
        self.assertEqual([alignment1, alignment2], readAlignments)


class TestBestAlignmentHSP(TestCase):
    """
    Test the L{dark.hits.bestAlignment} function when HSPs are used.
    """

    def testOneAlignment(self):
        """
        When one alignment is present that alignment must be returned by
        bestAlignment.
        """
        alignment = Alignment(44, "Seq 1")
        alignment.addHsp(HSP(10))
        alignment.addHsp(HSP(9))

        alignments = [alignment]
        hit = ReadAlignments(Read("id1", "aaa"), alignments)
        best = bestAlignment(hit)
        self.assertEqual("Seq 1", best.subjectTitle)
        self.assertEqual(44, best.subjectLength)

    def testThreeAlignments(self):
        """
        When three alignments are present, the one with the highest first HSP
        must be returned by bestAlignment.
        """
        alignment1 = Alignment(33, "Seq 1")
        alignment1.addHsp(HSP(10))
        alignment1.addHsp(HSP(9))

        alignment2 = Alignment(44, "Seq 2")
        alignment2.addHsp(HSP(30))
        alignment2.addHsp(HSP(29))

        alignment3 = Alignment(55, "Seq 3")
        alignment3.addHsp(HSP(20))
        alignment3.addHsp(HSP(19))

        alignments = [alignment1, alignment2, alignment3]
        hit = ReadAlignments(Read("id1", "aaa"), alignments)
        best = bestAlignment(hit)
        self.assertEqual("Seq 2", best.subjectTitle)
        self.assertEqual(44, best.subjectLength)


class TestBestAlignmentLSP(TestCase):
    """
    Test the L{dark.hits.bestAlignment} function when LSPs are used.
    """

    def testOneAlignment(self):
        """
        When one alignment is present that alignment must be returned by
        bestAlignment.
        """
        alignment = Alignment(44, "Seq 1")
        alignment.addHsp(LSP(10))
        alignment.addHsp(LSP(9))

        alignments = [alignment]
        readAlignments = ReadAlignments(Read("id0", "aaa"), alignments)
        best = bestAlignment(readAlignments)
        self.assertEqual("Seq 1", best.subjectTitle)
        self.assertEqual(44, best.subjectLength)

    def testThreeAlignments(self):
        """
        When three alignments are present, the one with the lowest first HSP
        must be returned by bestAlignment.
        """
        alignment1 = Alignment(33, "Seq 1")
        alignment1.addHsp(LSP(10))
        alignment1.addHsp(LSP(9))

        alignment2 = Alignment(44, "Seq 2")
        alignment2.addHsp(LSP(3))
        alignment2.addHsp(LSP(2))

        alignment3 = Alignment(55, "Seq 3")
        alignment3.addHsp(LSP(20))
        alignment3.addHsp(LSP(19))

        alignments = [alignment1, alignment2, alignment3]
        readAlignments = ReadAlignments(Read("id0", "aaa"), alignments)
        best = bestAlignment(readAlignments)
        self.assertEqual("Seq 2", best.subjectTitle)
        self.assertEqual(44, best.subjectLength)


class TestReadsAlignmentsParams(TestCase):
    """
    Test the L{dark.alignments.ReadsAlignmentsParams} class.
    """

    def testExpectedAttrs(self):
        """
        A ReadsAlignmentsParams instance must have the expected attributes.
        """
        applicationParams = {}
        params = ReadsAlignmentsParams(
            "application name", applicationParams, False, "Bit score"
        )
        self.assertEqual("application name", params.application)
        self.assertIs(applicationParams, params.applicationParams)
        self.assertFalse(params.subjectIsNucleotides)
        self.assertEqual("Bit score", params.scoreTitle)


class TestReadsAlignments(TestCase):
    """
    Test the L{dark.alignments.ReadsAlignments} class.
    """

    # NOTE: The ReadsAlignments class is a base class for concrete
    # implementations, such as BlastReadsAlignments. So it can only be
    # tested minimally by itself. For full(er) tests see
    # diamond/test_alignments.py and blast/test_alignments.py.

    def testExpectedAttrs(self):
        """
        A ReadsAlignments instance must have the expected attributes.
        """
        reads = Reads()
        params = {"application": "app name"}
        readsAlignments = ReadsAlignments(reads, params)
        self.assertIs(readsAlignments.reads, reads)
        self.assertEqual("app name", readsAlignments.params["application"])
        self.assertIs(params, readsAlignments.params)
        self.assertIs(HigherIsBetterScore, readsAlignments.scoreClass)

    def testNotIterable(self):
        """
        Iterating an empty ReadsAlignments must result in the empty list.
        """
        reads = Reads()
        readsAlignments = ReadsAlignments(reads, "applicationName", None)
        self.assertEqual([], list(readsAlignments))

    def testGetSubjectSequence(self):
        """
        A ReadsAlignments instance will not implement getSubjectSequence.
        Subclasses are expected to implement it.
        """
        reads = Reads()
        readsAlignments = ReadsAlignments(reads, "applicationName", None)
        error = "getSubjectSequence must be implemented by a subclass"
        six.assertRaisesRegex(
            self,
            NotImplementedError,
            error,
            readsAlignments.getSubjectSequence,
            "title",
        )



#!/usr/bin/env python

import sys
import csv

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA on stdin, write information about the "
            "number of various characters to stdout."
        )
    )

    parser.add_argument(
        "--ignoreCase",
        "-i",
        action="store_true",
        help="Map all characters to uppercase.",
    )

    parser.add_argument(
        "--sep", default="\t", help="The output field separator character."
    )

    parser.add_argument(
        "--omitIds",
        action="store_true",
        help="Do not print sequence ids as the first field",
    )

    parser.add_argument("--chars", "-c", required=True, help="The characters to count.")

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    chars = args.chars.upper() if args.ignoreCase else args.chars
    reads = parseFASTACommandLineOptions(args)

    writerow = csv.writer(sys.stdout, delimiter=args.sep).writerow

    for read in reads:
        length = len(read)
        sequence = read.sequence.upper() if args.ignoreCase else read.sequence
        for char in chars:
            count = sequence.count(char)
            out = [] if args.omitIds else [read.id]
            writerow(out + [char, count, length, f"{count / length:.4f}"])



import six
import bz2
from json import dumps, loads
from operator import itemgetter

from Bio.Blast import NCBIXML
from Bio.File import as_handle

from dark.hsp import HSP, LSP
from dark.score import HigherIsBetterScore
from dark.alignments import Alignment, ReadAlignments
from dark.blast.hsp import normalizeHSP


class XMLRecordsReader:
    """
    Provide a method that yields parsed XML records from a file. Store and
    make accessible the global BLAST parameters.

    @ivar params: A C{dict} of global BLAST parameters.
    @param filename: A C{str} filename or an open file pointer, containing XML
        BLAST records.
    """

    def __init__(self, filename):
        self._filename = filename
        self.params = None  # Set below, in records.

    def _convertBlastRecordToDict(self, record):
        """
        Pull (only) the fields we use out of the record and return them as a
        dict.  Although we take the title from each alignment description, we
        save space in the JSON output by storing it in the alignment dict (not
        in a separated 'description' dict). When we undo this conversion (in
        JSONRecordsReader._convertDictToBlastRecord) we'll pull the title out
        of the alignment dict and put it into the right place in the BLAST
        record.

        @param record: An instance of C{Bio.Blast.Record.Blast}. The attributes
            on this don't seem to be documented. You'll need to look at the
            BioPython source to see everything it contains.
        @return: A C{dict} with 'alignments' and 'query' keys.
        """
        alignments = []
        for alignment in record.alignments:
            hsps = []
            for hsp in alignment.hsps:
                hsps.append(
                    {
                        "bits": hsp.bits,
                        "expect": hsp.expect,
                        "frame": hsp.frame,
                        "identicalCount": hsp.identities,
                        "positiveCount": hsp.positives,
                        "query": hsp.query,
                        "query_start": hsp.query_start,
                        "query_end": hsp.query_end,
                        "sbjct": hsp.sbjct,
                        "sbjct_start": hsp.sbjct_start,
                        "sbjct_end": hsp.sbjct_end,
                    }
                )

            alignments.append(
                {
                    "hsps": hsps,
                    "length": alignment.length,
                    "title": alignment.title,
                }
            )

        return {
            "alignments": alignments,
            "query": record.query,
        }

    def _convertBlastParamsToDict(self, record):
        """
        Pull the global BLAST parameters out of a BLAST record and return
        them as a C{dict}.

        Some of these attributes are useless (not filled in), but we record
        them all just in case we one day need them or they start to be used or
        they disappear etc. Any of those changes might alert us that something
        has changed in BLAST XML output or in BioPython.

        @param record: An instance of C{Bio.Blast.Record.Blast}. The attributes
            on this don't seem to be documented. You'll need to look at the
            BioPython source to see everything it contains.
        @return: A C{dict}, as described above.
        """
        result = {}
        for attr in (
            # From Bio.Blast.Record.Header
            "application",
            "version",
            "date",
            "reference",
            "query",
            "query_letters",
            "database",
            "database_sequences",
            "database_letters",
            # From Bio.Blast.Record.DatabaseReport
            "database_name",
            "posted_date",
            "num_letters_in_database",
            "num_sequences_in_database",
            "ka_params",
            "gapped",
            "ka_params_gap",
            # From Bio.Blast.Record.Parameters
            "matrix",
            "gap_penalties",
            "sc_match",
            "sc_mismatch",
            "num_hits",
            "num_sequences",
            "num_good_extends",
            "num_seqs_better_e",
            "hsps_no_gap",
            "hsps_prelim_gapped",
            "hsps_prelim_gapped_attemped",
            "hsps_gapped",
            "query_id",
            "query_length",
            "database_length",
            "effective_hsp_length",
            "effective_query_length",
            "effective_database_length",
            "effective_search_space",
            "effective_search_space_used",
            "frameshift",
            "threshold",
            "window_size",
            "dropoff_1st_pass",
            "gap_x_dropoff",
            "gap_x_dropoff_final",
            "gap_trigger",
            "blast_cutoff",
        ):
            result[attr] = getattr(record, attr)
        return result

    def records(self):
        """
        Yield BLAST records, as read by the BioPython NCBIXML.parse
        method. Set self.params from data in the first record.
        """
        first = True
        with as_handle(self._filename) as fp:
            for record in NCBIXML.parse(fp):
                if first:
                    self.params = self._convertBlastParamsToDict(record)
                    first = False
                yield record

    def saveAsJSON(self, fp):
        """
        Write the records out as JSON. The first JSON object saved contains
        the BLAST parameters.

        @param fp: A C{str} file pointer to write to.
        """
        first = True
        for record in self.records():
            if first:
                print(dumps(self.params, separators=(",", ":")), file=fp)
                first = False
            print(
                dumps(self._convertBlastRecordToDict(record), separators=(",", ":")),
                file=fp,
            )


class JSONRecordsReader:
    """
    Provide a method that yields JSON records from a file. Store, check, and
    make accessible the global BLAST parameters.

    @param filename: A C{str} filename containing JSON BLAST records.
    @param scoreClass: A class to hold and compare scores (see scores.py).
        Default is C{HigherIsBetterScore}, for comparing bit scores. If you
        are using e-values, pass LowerIsBetterScore instead.
    """

    # Note that self._fp is opened in self.__init__, accessed in
    # self._params and in self.records, and closed in self.close.

    def __init__(self, filename, scoreClass=HigherIsBetterScore):
        self._filename = filename
        self._scoreClass = scoreClass
        if scoreClass is HigherIsBetterScore:
            self._hspClass = HSP
        else:
            self._hspClass = LSP

        self._open(filename)
        self.application = self.params["application"].lower()

    def _open(self, filename):
        """
        Open the input file. Set self._fp to point to it. Read the first
        line of parameters.

        @param filename: A C{str} filename containing JSON BLAST records.
        @raise ValueError: if the first line of the file isn't valid JSON,
            if the input file is empty, or if the JSON does not contain an
            'application' key.
        """
        if filename.endswith(".bz2"):
            if six.PY3:
                self._fp = bz2.open(filename, mode="rt", encoding="UTF-8")
            else:
                self._fp = bz2.BZ2File(filename)
        else:
            self._fp = open(filename)

        line = self._fp.readline()
        if not line:
            raise ValueError("JSON file %r was empty." % self._filename)

        try:
            self.params = loads(line[:-1])
        except ValueError as e:
            raise ValueError(
                "Could not convert first line of %r to JSON (%s). "
                "Line is %r." % (self._filename, e, line[:-1])
            )
        else:
            if "application" not in self.params:
                raise ValueError(
                    "%r appears to be an old JSON file with no BLAST global "
                    "parameters. Please re-run convert-blast-xml-to-json.py "
                    "to convert it to the newest format." % self._filename
                )

    def _dictToAlignments(self, blastDict, read):
        """
        Take a dict (made by XMLRecordsReader._convertBlastRecordToDict)
        and convert it to a list of alignments.

        @param blastDict: A C{dict}, from convertBlastRecordToDict.
        @param read: A C{Read} instance, containing the read that BLAST used
            to create this record.
        @raise ValueError: If the query id in the BLAST dictionary does not
            match the id of the read.
        @return: A C{list} of L{dark.alignment.Alignment} instances.
        """
        if blastDict["query"] != read.id and blastDict["query"].split()[0] != read.id:
            raise ValueError(
                "The reads you have provided do not match the BLAST output: "
                "BLAST record query id (%s) does not match the id of the "
                "supposedly corresponding read (%s)." % (blastDict["query"], read.id)
            )

        alignments = []
        getScore = itemgetter("bits" if self._hspClass is HSP else "expect")

        for blastAlignment in blastDict["alignments"]:
            alignment = Alignment(blastAlignment["length"], blastAlignment["title"])
            alignments.append(alignment)
            for blastHsp in blastAlignment["hsps"]:
                score = getScore(blastHsp)
                normalized = normalizeHSP(blastHsp, len(read), self.application)
                hsp = self._hspClass(
                    score,
                    readStart=normalized["readStart"],
                    readEnd=normalized["readEnd"],
                    readStartInSubject=normalized["readStartInSubject"],
                    readEndInSubject=normalized["readEndInSubject"],
                    readFrame=blastHsp["frame"][0],
                    subjectStart=normalized["subjectStart"],
                    subjectEnd=normalized["subjectEnd"],
                    subjectFrame=blastHsp["frame"][1],
                    readMatchedSequence=blastHsp["query"],
                    subjectMatchedSequence=blastHsp["sbjct"],
                    # Use blastHsp.get on identicalCount and positiveCount
                    # because they were added in version 2.0.3 and will not
                    # be present in any of our JSON output generated before
                    # that. Those values will be None for those JSON files,
                    # but that's much better than no longer being able to
                    # read all that data.
                    identicalCount=blastHsp.get("identicalCount"),
                    positiveCount=blastHsp.get("positiveCount"),
                )

                alignment.addHsp(hsp)

        return alignments

    def readAlignments(self, reads):
        """
        Read lines of JSON from self._filename, convert them to read alignments
        and yield them.

        @param reads: An iterable of L{Read} instances, corresponding to the
            reads that were given to BLAST.
        @raise ValueError: If any of the lines in the file cannot be converted
            to JSON.
        @return: A generator that yields C{dark.alignments.ReadAlignments}
            instances.
        """
        if self._fp is None:
            self._open(self._filename)

        reads = iter(reads)

        try:
            for lineNumber, line in enumerate(self._fp, start=2):
                try:
                    record = loads(line[:-1])
                except ValueError as e:
                    raise ValueError(
                        "Could not convert line %d of %r to JSON (%s). "
                        "Line is %r." % (lineNumber, self._filename, e, line[:-1])
                    )
                else:
                    try:
                        read = next(reads)
                    except StopIteration:
                        raise ValueError(
                            "Read generator failed to yield read number %d "
                            "during parsing of BLAST file %r."
                            % (lineNumber - 1, self._filename)
                        )
                    else:
                        alignments = self._dictToAlignments(record, read)
                        yield ReadAlignments(read, alignments)
        finally:
            self._fp.close()
            self._fp = None



#!/usr/bin/env python

"""
Read a FASTA or FASTQ file (or read stdin) and print a line for each sequence,
with the length of the sequence followed by its name.
"""

import sys
import argparse
from Bio import SeqIO


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Print lengths of reads in a FASTA or FASTQ file",
    )

    parser.add_argument(
        "reads",
        nargs="?",
        type=argparse.FileType("r"),
        default=sys.stdin,
        help="The input sequence file (in either FASTA or FASTQ format). "
        "If not given, sequences will be read from stdin.",
    )

    parser.add_argument(
        "--format",
        default="fasta",
        choices=["fasta", "fastq"],
        help="The format of the input.",
    )

    args = parser.parse_args()

    for seq in SeqIO.parse(args.reads, args.format):
        print("%d %s" % (len(seq), seq.description))



import sys

from dark.btop import countGaps


def _debugPrint(hsp, queryLen, localDict, msg=""):
    """
    Print debugging information showing the local variables used during
    a call to normalizeHSP and the hsp and then raise an C{AssertionError}.

    @param hsp: The HSP C{dict} passed to normalizeHSP.
    @param queryLen: the length of the query sequence.
    @param localDict: A C{dict} of local variables (as produced by locals()).
    @param msg: A C{str} message to raise C{AssertionError} with.
    @raise AssertionError: unconditionally.
    """
    print("normalizeHSP error:", file=sys.stderr)
    print("  queryLen: %d" % queryLen, file=sys.stderr)

    print("  Original HSP:", file=sys.stderr)
    for attr in [
        "bits",
        "btop",
        "expect",
        "frame",
        "query_end",
        "query_start",
        "sbjct",
        "query",
        "sbjct_end",
        "sbjct_start",
    ]:
        print("    %s: %r" % (attr, hsp[attr]), file=sys.stderr)

    print("  Local variables:", file=sys.stderr)
    for var in sorted(localDict):
        if var != "hsp":
            print("    %s: %s" % (var, localDict[var]), file=sys.stderr)

    raise AssertionError(msg)


def _sanityCheck(
    subjectStart,
    subjectEnd,
    queryStart,
    queryEnd,
    queryStartInSubject,
    queryEndInSubject,
    hsp,
    queryLen,
    subjectGaps,
    queryGaps,
    localDict,
):
    """
    Perform some sanity checks on an HSP. Call _debugPrint on any error.

    @param subjectStart: The 0-based C{int} start offset of the match in the
        subject.
    @param subjectEnd: The 0-based C{int} end offset of the match in the
        subject.
    @param queryStart: The 0-based C{int} start offset of the match in the
        query.
    @param queryEnd: The 0-based C{int} end offset of the match in the query.
    @param queryStartInSubject: The 0-based C{int} offset of where the query
        starts in the subject.
    @param queryEndInSubject: The 0-based C{int} offset of where the query
        ends in the subject.
    @param hsp: The HSP C{dict} passed to normalizeHSP.
    @param queryLen: the C{int} length of the query sequence.
    @param subjectGaps: the C{int} number of gaps in the subject.
    @param queryGaps: the C{int} number of gaps in the query.
    @param localDict: A C{dict} of local variables from our caller (as
        produced by locals()).
    """
    # Subject indices must always be ascending.
    if subjectStart >= subjectEnd:
        _debugPrint(hsp, queryLen, localDict, "subjectStart >= subjectEnd")

    subjectMatchLength = subjectEnd - subjectStart
    queryMatchLength = queryEnd - queryStart

    # Sanity check that the length of the matches in the subject and query
    # are identical, taking into account gaps in both.
    subjectMatchLengthWithGaps = subjectMatchLength + subjectGaps
    queryMatchLengthWithGaps = queryMatchLength + queryGaps
    if subjectMatchLengthWithGaps != queryMatchLengthWithGaps:
        _debugPrint(
            hsp,
            queryLen,
            localDict,
            "Including gaps, subject match length (%d) != Query match "
            "length (%d)" % (subjectMatchLengthWithGaps, queryMatchLengthWithGaps),
        )

    if queryStartInSubject > subjectStart:
        _debugPrint(
            hsp,
            queryLen,
            localDict,
            "queryStartInSubject (%d) > subjectStart (%d)"
            % (queryStartInSubject, subjectStart),
        )
    if queryEndInSubject < subjectEnd:
        _debugPrint(
            hsp,
            queryLen,
            localDict,
            "queryEndInSubject (%d) < subjectEnd (%d)"
            % (queryEndInSubject, subjectEnd),
        )


def normalizeHSP(hsp, queryLen, diamondTask):
    """
    Examine an HSP and return information about where the query and subject
    match begins and ends.  Return a dict with keys that allow the query to
    be displayed against the subject. The returned readStartInSubject and
    readEndInSubject indices are offsets into the subject. I.e., they
    indicate where in the subject the query falls.

    In the returned object, all indices are suitable for Python string
    slicing etc.  We must be careful to convert from the 1-based offsets
    found in DIAMOND output properly.

    hsp['frame'] is a value from {-3, -2, -1, 1, 2, 3}. The sign indicates
    negative or positive sense (i.e., the direction of reading through the
    query to get the alignment). The frame value is the nucleotide match offset
    modulo 3, plus one (i.e., it tells us which of the 3 possible query reading
    frames was used in the match).

    NOTE: the returned readStartInSubject value may be negative.  We consider
    the subject sequence to start at offset 0.  So if the query string has
    sufficient additional nucleotides before the start of the alignment
    match, it may protrude to the left of the subject. Similarly, the returned
    readEndInSubject can be greater than the subjectEnd.

    @param hsp: an HSP in the form of a C{dict}, built from a DIAMOND record.
        All passed offsets are 1-based.
    @param queryLen: the length of the query sequence.
    @param diamondTask: The C{str} command-line matching algorithm that was
        run (either 'blastx' or 'blastp').
    @return: A C{dict} with C{str} keys and C{int} offset values. Keys are
            readStart
            readEnd
            readStartInSubject
            readEndInSubject
            subjectStart
            subjectEnd
        The returned offset values are all zero-based.
    """

    queryGaps, subjectGaps = countGaps(hsp["btop"])

    # Make some variables using Python's standard string indexing (start
    # offset included, end offset not). No calculations in this function
    # are done with the original 1-based HSP variables.
    queryStart = hsp["query_start"] - 1
    queryEnd = hsp["query_end"]
    subjectStart = hsp["sbjct_start"] - 1
    subjectEnd = hsp["sbjct_end"]

    queryReversed = hsp["frame"] < 0

    # Query offsets must be ascending, unless we're looking at blastx output
    # and the query was reversed for the match.
    if queryStart >= queryEnd:
        if diamondTask == "blastx" and queryReversed:
            # Compute new query start and end indices, based on their
            # distance from the end of the string.
            #
            # Above we took one off the start index, so we need to undo
            # that (because the start is actually the end). We didn't take
            # one off the end index, and need to do that now (because the
            # end is actually the start).
            queryStart = queryLen - (queryStart + 1)
            queryEnd = queryLen - (queryEnd - 1)
        else:
            _debugPrint(hsp, queryLen, locals(), "queryStart >= queryEnd")

    if diamondTask == "blastx":
        # In DIAMOND blastx output, subject offsets are based on protein
        # sequence length but queries (and the reported offsets) are
        # nucleotide.  Convert the query offsets to protein because we will
        # plot against the subject (protein).
        #
        # Convert queryLen and the query nucleotide start and end offsets
        # to be valid for the query after translation to AAs. When
        # translating, DIAMOND may ignore some nucleotides at the start
        # and/or the end of the original DNA query. At the start this is
        # due to the frame in use, and at the end it is due to always using
        # three nucleotides at a time to form codons.
        #
        # So, for example, a query of 6 nucleotides that is translated in
        # frame 2 (i.e., the translation starts from the second nucleotide)
        # will have length 1 as an AA sequence. The first nucleotide is
        # ignored due to the frame and the last two due to there not being
        # enough final nucleotides to make another codon.
        #
        # In the following, the subtraction accounts for the first form of
        # loss and the integer division for the second.
        initiallyIgnored = abs(hsp["frame"]) - 1
        queryLen = (queryLen - initiallyIgnored) // 3
        queryStart = (queryStart - initiallyIgnored) // 3
        queryEnd = (queryEnd - initiallyIgnored) // 3

    # unmatchedQueryLeft is the number of query bases that will extend
    # to the left of the start of the subject in our plots.
    unmatchedQueryLeft = queryStart

    # Set the query offsets into the subject.
    queryStartInSubject = subjectStart - unmatchedQueryLeft
    queryEndInSubject = queryStartInSubject + queryLen + queryGaps

    _sanityCheck(
        subjectStart,
        subjectEnd,
        queryStart,
        queryEnd,
        queryStartInSubject,
        queryEndInSubject,
        hsp,
        queryLen,
        subjectGaps,
        queryGaps,
        locals(),
    )

    return {
        "readStart": queryStart,
        "readEnd": queryEnd,
        "readStartInSubject": queryStartInSubject,
        "readEndInSubject": queryEndInSubject,
        "subjectStart": subjectStart,
        "subjectEnd": subjectEnd,
    }



from six.moves import builtins
from unittest import TestCase
from unittest.mock import patch, mock_open

from dark.filter import ReadSetFilter, TitleFilter
from dark.reads import Read
from dark.titles import TitleAlignment, TitleAlignments


class TitleFilterTest(TestCase):
    """
    Tests for the L{dark.filter.TitleFilter} class.
    """

    def testNoRestriction(self):
        """
        Testing for acceptance against a title filter that has no
        restrictions should return C{TitleFilter.DEFAULT_ACCEPT}.
        """
        tf = TitleFilter()
        self.assertEqual(TitleFilter.DEFAULT_ACCEPT, tf.accept("hey"))

    def testPositiveRegex(self):
        """
        Testing for acceptance against a title filter with a positive regex
        must work.
        """
        tf = TitleFilter(positiveRegex=r"x+\s")
        self.assertEqual(TitleFilter.DEFAULT_ACCEPT, tf.accept("hey xxx you"))
        self.assertEqual(TitleFilter.REJECT, tf.accept("hey xxyou"))

    def testNegativeRegex(self):
        """
        Testing for acceptance against a title filter with a negative regex
        must work.
        """
        tf = TitleFilter(negativeRegex=r"x+\s")
        self.assertEqual(TitleFilter.REJECT, tf.accept("hey xxx you"))
        self.assertEqual(TitleFilter.DEFAULT_ACCEPT, tf.accept("hey xxyou"))

    def testPositiveRegexHasPrecedenceOverRepeatedTruncatedTitle(self):
        """
        Testing for acceptance against a title filter with a positive regex
        must have precedence over checking for truncated titles when the same
        non-matching title (that will be truncated) is passed twice.
        """
        tf = TitleFilter(positiveRegex=r"xxxxx", truncateAfter="virus")
        self.assertEqual(TitleFilter.REJECT, tf.accept("spotty virus 1"))
        self.assertEqual(TitleFilter.REJECT, tf.accept("spotty virus 1"))

    def testNegativeRegexHasPrecedenceOverRepeatedTruncatedTitle(self):
        """
        Testing for acceptance against a title filter with a negative regex
        must have precedence over checking for truncated titles when the same
        matching title (that will be truncated) is passed twice.
        """
        tf = TitleFilter(negativeRegex=r"spotty", truncateAfter="virus")
        self.assertEqual(TitleFilter.REJECT, tf.accept("spotty virus 1"))
        self.assertEqual(TitleFilter.REJECT, tf.accept("spotty virus 1"))

    def testFullWordTruncation(self):
        """
        Testing for acceptance against a title filter with title truncation
        in effect must work if the title contains the C{truncateAfter} string
        as a distint word.
        """
        tf = TitleFilter(truncateAfter=r"virus")
        # Note that the truncation code will chop off the first part of the
        # title (the title ID).
        self.assertEqual(
            TitleFilter.DEFAULT_ACCEPT,
            tf.accept("gi|400684|gb|AY421767.1| herpes virus 1"),
        )
        self.assertEqual(
            TitleFilter.REJECT, tf.accept("gi|400684|gb|AY421767.1| herpes virus 2")
        )

    def testPartialWordTruncation(self):
        """
        Testing for acceptance against a title filter with title truncation
        in effect must work if the title contains the C{truncateAfter} string
        as a partial word.
        """
        tf = TitleFilter(truncateAfter=r"virus")
        # Note that the truncation code will chop off the first part of the
        # title (the title ID).
        self.assertEqual(
            TitleFilter.DEFAULT_ACCEPT,
            tf.accept("gi|400684|gb|AY421767.1| rotavirus 1"),
        )
        self.assertEqual(
            TitleFilter.REJECT, tf.accept("gi|400684|gb|AY421767.1| rotavirus 2")
        )

    def testWordTruncationRepeat(self):
        """
        Testing for acceptance against a title filter with title truncation
        in effect must allow the exact same title twice, even if the title
        is being truncated.
        """
        tf = TitleFilter(truncateAfter=r"virus")
        # Note that the truncation code will chop off the first part of the
        # title (the title ID).
        self.assertEqual(
            TitleFilter.DEFAULT_ACCEPT,
            tf.accept("gi|400684|gb|AY421767.1| herpes virus 1"),
        )
        self.assertEqual(
            TitleFilter.DEFAULT_ACCEPT,
            tf.accept("gi|400684|gb|AY421767.1| herpes virus 1"),
        )

    def testWhitelist(self):
        """
        Testing for acceptance against a title filter with a whitelist
        must work even when a title is ruled out for other violations.
        """
        tf = TitleFilter(whitelist=["always ok"], negativeRegex="ok")
        self.assertEqual(TitleFilter.WHITELIST_ACCEPT, tf.accept("always ok"))
        self.assertEqual(TitleFilter.REJECT, tf.accept("always ok not"))

    def testBlacklist(self):
        """
        Testing for acceptance against a title filter with a blacklist
        must work.
        """
        tf = TitleFilter(blacklist=["never ok"], positiveRegex="ok")
        self.assertEqual(TitleFilter.REJECT, tf.accept("never ok"))

    def testBlacklistFile(self):
        """
        Testing for acceptance against a title filter with a blacklist file.
        """
        data = "\n".join(["id1", "id2"]) + "\n"
        with patch.object(builtins, "open", mock_open(read_data=data)):
            tf = TitleFilter(blacklistFile="black.txt")
            self.assertEqual(TitleFilter.REJECT, tf.accept("id1"))
            self.assertEqual(TitleFilter.REJECT, tf.accept("id2"))
            self.assertEqual(TitleFilter.DEFAULT_ACCEPT, tf.accept("id3"))

    def testBlacklistFileAndBlacklist(self):
        """
        Testing for acceptance against a title filter with a blacklist file and
        some specific other blacklist titles.
        """
        data = "\n".join(["id1", "id2"]) + "\n"
        with patch.object(builtins, "open", mock_open(read_data=data)):
            tf = TitleFilter(blacklistFile="black.txt", blacklist=set(["id3"]))
            self.assertEqual(TitleFilter.REJECT, tf.accept("id1"))
            self.assertEqual(TitleFilter.REJECT, tf.accept("id2"))
            self.assertEqual(TitleFilter.REJECT, tf.accept("id3"))
            self.assertEqual(TitleFilter.DEFAULT_ACCEPT, tf.accept("id4"))

    def testWhitelistTakesPrecedenceOverBlacklist(self):
        """
        Testing for acceptance against a title filter with a whitelist
        and a blacklist that contain the same title must work (the whitelist
        takes precedence).
        """
        tf = TitleFilter(whitelist=["always ok"], blacklist=["always ok"])
        self.assertEqual(TitleFilter.WHITELIST_ACCEPT, tf.accept("always ok"))

    def testWhitelistOnly(self):
        """
        Testing for acceptance against a title filter with a whitelist
        and a negative regex that matches everything.
        """
        tf = TitleFilter(whitelist=["always ok"], negativeRegex=".")
        self.assertEqual(TitleFilter.WHITELIST_ACCEPT, tf.accept("always ok"))
        self.assertEqual(TitleFilter.REJECT, tf.accept("always not ok"))
        self.assertEqual(TitleFilter.REJECT, tf.accept("rubbish"))

    def testWhitelistFileOnly(self):
        """
        Testing for acceptance against a title filter with a whitelist file
        and a negative regex that matches everything.
        """
        data = "\n".join(["id1", "id2"]) + "\n"
        with patch.object(builtins, "open", mock_open(read_data=data)):
            tf = TitleFilter(whitelistFile="white.txt", negativeRegex=".")
            self.assertEqual(TitleFilter.WHITELIST_ACCEPT, tf.accept("id1"))
            self.assertEqual(TitleFilter.WHITELIST_ACCEPT, tf.accept("id2"))
            self.assertEqual(TitleFilter.REJECT, tf.accept("id3"))

    def testWhitelistFileAndWhitelistOnly(self):
        """
        Testing for acceptance against a title filter with a whitelist file
        and some specific whitelist titles, with a negative regex that matches
        everything.
        """
        data = "\n".join(["id1", "id2"]) + "\n"
        with patch.object(builtins, "open", mock_open(read_data=data)):
            tf = TitleFilter(
                whitelistFile="white.txt", whitelist=set(["id3"]), negativeRegex="."
            )
            self.assertEqual(TitleFilter.WHITELIST_ACCEPT, tf.accept("id1"))
            self.assertEqual(TitleFilter.WHITELIST_ACCEPT, tf.accept("id2"))
            self.assertEqual(TitleFilter.WHITELIST_ACCEPT, tf.accept("id3"))
            self.assertEqual(TitleFilter.REJECT, tf.accept("id4"))


class ReadSetTest(TestCase):
    """
    Tests for the L{dark.filter.ReadSetFilter} class.
    """

    def makeTitleAlignments(self, *readIds):
        """
        Create a TitleAlignments instance containing reads with the
        ids given by C{ids}.

        param readIds: A C{list} of integer ids for reads.
        @return: A C{TitleAlignments} instance with reads with the given ids.
        """
        titleAlignments = TitleAlignments("subject title", 55)
        for readId in readIds:
            titleAlignment = TitleAlignment(Read("id" + str(readId), "A"), [])
            titleAlignments.addAlignment(titleAlignment)
        return titleAlignments

    def testFirstUse(self):
        """
        Testing for acceptance against a read set filter that has not been
        used should return C{True}.
        """
        titleAlignments = self.makeTitleAlignments()
        rsf = ReadSetFilter(0.9)
        self.assertTrue(rsf.accept("title1", titleAlignments))

    def testDuplicateSingleRead(self):
        """
        Testing for acceptance against a read set filter that has already
        seen the exact set should return C{False} if the C{minNew} threshold
        is non-zero.
        """
        rsf = ReadSetFilter(0.9)
        rsf.accept("title1", self.makeTitleAlignments(0))
        self.assertFalse(rsf.accept("title2", self.makeTitleAlignments(0)))

    def testDuplicateSingleReadZeroThreshold(self):
        """
        Testing for acceptance against a read set filter that has already
        seen the exact set should return C{True} if the C{minNew} threshold
        is zero.
        """
        rsf = ReadSetFilter(0.0)
        rsf.accept("title1", self.makeTitleAlignments(0))
        self.assertTrue(rsf.accept("title2", self.makeTitleAlignments(0)))

    def testDifferentSet(self):
        """
        Testing for acceptance against a read set filter that has seen a set
        should return C{True} if the new set is totally different.
        """
        rsf = ReadSetFilter(1.0)
        rsf.accept("title1", self.makeTitleAlignments(0))
        self.assertTrue(rsf.accept("title2", self.makeTitleAlignments(1)))

    def testSufficientlyDifferent(self):
        """
        Testing for acceptance against a read set filter that has seen several
        sets should return C{True} if the new set is sufficiently different.
        """
        rsf = ReadSetFilter(0.5)
        rsf.accept("title1", self.makeTitleAlignments(0, 1, 2, 3, 4))
        rsf.accept("title2", self.makeTitleAlignments(5, 6, 7, 8, 9))
        self.assertTrue(
            rsf.accept("title3", self.makeTitleAlignments(0, 1, 2, 5, 6, 7))
        )

    def testInsufficientlyDifferent(self):
        """
        Testing for acceptance against a read set filter that has seen several
        sets should return C{False} if the new set is insufficiently different.
        """
        rsf = ReadSetFilter(0.5)
        rsf.accept("title1", self.makeTitleAlignments(0, 1, 2, 3, 4))
        rsf.accept("title2", self.makeTitleAlignments(5, 6, 7, 8, 9))
        self.assertFalse(rsf.accept("title3", self.makeTitleAlignments(0, 1, 2, 11)))

    def testThresholdRoundsUp(self):
        """
        Testing for acceptance should round up the needed number of new reads.
        """
        rsf = ReadSetFilter(0.5)
        rsf.accept("title1", self.makeTitleAlignments(0, 1, 2, 3, 4))
        # If we pass a read set of size three, two of the reads will need to be
        # different.
        self.assertFalse(rsf.accept("title2", self.makeTitleAlignments(0, 1, 6)))

    def testRepeatTitle(self):
        """
        Testing for acceptance on a title that has been seen before (in an
        accepted read set) must raise C{AssertionError}.
        """
        rsf = ReadSetFilter(0.5)
        rsf.accept("title1", self.makeTitleAlignments(0, 1, 2, 3, 4))
        self.assertRaises(
            AssertionError, rsf.accept, "title1", self.makeTitleAlignments()
        )

    def testInvalidates(self):
        """
        It must be possible to retrieve the list of titles that were
        invalidated by an earlier title's read set.
        """
        rsf = ReadSetFilter(0.5)
        rsf.accept("title1", self.makeTitleAlignments(0))
        rsf.accept("title2", self.makeTitleAlignments(0))
        rsf.accept("title3", self.makeTitleAlignments(1))
        rsf.accept("title4", self.makeTitleAlignments(0))
        self.assertEqual(["title2", "title4"], rsf.invalidates("title1"))

    def testInvalidatesEmpty(self):
        """
        The list of titles invalidated by an earlier title that didn't
        invalidate anything must be empty.
        """
        rsf = ReadSetFilter(0.5)
        self.assertEqual([], rsf.invalidates("title1"))


class FakeCursor:
    def __init__(self, results):
        self._results = results
        self._index = -1

    def execute(self, p):
        pass

    def fetchone(self):
        self._index += 1
        return self._results[self._index]

    def close(self):
        pass


class FakeDbConnection:
    """
    FakeDbConnection and FakeCursor fake results
    for database calls.
    """

    def __init__(self, results):
        self._results = results
        self.open = True

    def cursor(self):
        return FakeCursor(self._results)

    def close(self):
        self.open = False



#!/usr/bin/env python

from Bio import SeqIO
import sys

if len(sys.argv) < 2:
    print("Usage: %s file.sff > file.fastq" % sys.argv[0], file=sys.stderr)
    sys.exit(1)
else:
    SeqIO.convert(sys.argv[1], "sff", sys.stdout, "fastq")



import six

from tempfile import mkdtemp
from shutil import rmtree
from os.path import join
from subprocess import CalledProcessError

from dark.diamond.conversion import FIELDS, DiamondTabularFormat
from dark.process import Executor
from dark.utils import cd


def diamondInstalled():
    """
    Test if DIAMOND is installed.

    @return: A C{bool}, which is C{True} if DIAMOND seems to be installed.
    """
    try:
        Executor().execute("diamond help")
    except CalledProcessError:
        return False
    else:
        return True


class DiamondExecutor:
    """

    @param dryRun: If C{True} do not actually execute the DIAMOND commands.
    """

    SUBJECTS_FILENAME = "subjects.fasta"
    QUERIES_FILENAME = "queries.fasta"
    OUTPUT_FILENAME = "diamond.tsv"

    def __init__(self, dryRun=False):
        self._dirty = False
        self._dir = mkdtemp()
        self._subjectsFp = None
        self._subjectsExist = False
        self._executor = Executor(dryRun)

    def addSubject(self, subject):
        """
        Add a subject sequence to the database.

        @param subject: A C{dark.reads.Read} instance.
        """
        if self._subjectsFp is None:
            if six.PY3:
                self._subjectsFp = open(
                    join(self._dir, self.SUBJECTS_FILENAME), "a", encoding="utf-8"
                )
            else:
                self._subjectsFp = open(join(self._dir, self.SUBJECTS_FILENAME), "a")

        print(subject.toString("fasta"), end="", file=self._subjectsFp)
        self._subjectsExist = self._dirty = True

    def cleanup(self):
        """
        Remove the temporary directory we made.
        """
        if self._subjectsFp:
            self._subjectsFp.close()
            self._subjectsFp = None
        rmtree(self._dir)

    def search(self, reads, fieldNames=None):
        """
        Match reads against the database.

        @param reads: An instance of C{dark.reads.Reads}.
        @param fieldNames: An iterable of C{str} field names for DIAMOND
            tabular output (format 6). See diamond help for the names of all
            available fields.
        @return: A generator that yields C{dict}s with keys as in
            C{fieldNames}.
        """
        if not self._subjectsExist:
            raise ValueError("No subject sequences in the database")

        with cd(self._dir):
            if self._dirty:
                self._subjectsFp.close()
                self._subjectsFp = None
                self._executor.execute(
                    "diamond makedb --db database --in %s" % self.SUBJECTS_FILENAME
                )

            with open(self.QUERIES_FILENAME, "w") as fp:
                count = reads.save(fp, format_="fastq")

            if count == 0:
                raise ValueError("No query sequences were passed")

            fieldNames = fieldNames or FIELDS.split()

            self._executor.execute(
                "diamond blastx --db database --query %s --outfmt 6 %s > %s"
                % (self.QUERIES_FILENAME, " ".join(fieldNames), self.OUTPUT_FILENAME)
            )

            dtf = DiamondTabularFormat(fieldNames)

            for diamondDict in dtf.diamondTabularFormatToDicts(self.OUTPUT_FILENAME):
                yield diamondDict

    def __enter__(self):
        return self

    def __exit__(self, excType, excValue, traceback):
        self.cleanup()
